from .database import *
from .dataset.data_blending import DataSetBlender
from .dataset.fields import (
    DataSetFilterException,
    DataType,
    Field,
)
from .dataset.intervals import (
    NumericInterval,
    day,
    hour,
    month,
    quarter,
    week,
    year,
)
from .dataset.joins import Join
from .dataset.klass import DataSet
from .dataset.modifiers import (
    OmitFromRollup,
    Rollup,
    ResultSet,
)
from .dataset.operations import (
    CumMean,
    CumProd,
    CumSum,
    Operation,
    RollingMean,
    Share,
)
from .dataset.references import (
    DayOverDay,
    MonthOverMonth,
    QuarterOverQuarter,
    WeekOverWeek,
    YearOverYear,
)
from .exceptions import DataSetException
from .widgets import *

__version__ = "7.2.5"



from copy import deepcopy

from pypika import (
    Case,
    terms,
)
from pypika.terms import BasicCriterion
from typing import (
    Tuple,
    List,
)

from fireant.dataset.fields import (
    DataType,
    Field,
)
from fireant.dataset.modifiers import (
    DimensionModifier,
    ResultSet,
)
from ..utils import (
    alias_selector,
    flatten,
)


def _is_metric(field: Field) -> bool:
    """
    Returns whether a field is a metric.

    :param field: A Field instance.
    :return: A boolean.
    """
    # Only terms wrapped with aggregate functions, such as SUM, will evaluate is_aggregate to True in Pypika.
    # This is a good way to validate that the set dimension, in question, is actually encompassing a metric.
    return field.is_aggregate


def _make_set_dimension(set_filter: Field, target_dataset: 'DataSet') -> Field:
    """
    Returns a new dimension that uses a CASE statement as its definition, in order to represent membership to a set,
    given the provided conditional.

    :param set_filter: A ResultSet instance.
    :param target_dataset: A DataSet instance, that will be used as the dataset for which the new dimension
                           will be applied to.
    :return: A new Field instance.
    """
    old_definition = set_filter.filter

    while hasattr(old_definition, 'definition'):
        old_definition = old_definition.definition

    old_definition = deepcopy(old_definition)
    old_definition_sql = old_definition.get_sql(quote_char="")

    set_dimension = deepcopy(set_filter.filter.field)

    is_metric = _is_metric(set_dimension)

    # When using data blending, the dataset table of the set filter needs to be re-mapped to the table in the
    # target dataset (i.e. primary or secondary). The easiest way to do that is to select the field in the
    # target dataset directly.
    if target_dataset and not is_metric and isinstance(old_definition, (
        terms.ContainsCriterion, terms.NullCriterion,
        terms.BetweenCriterion, terms.BitwiseAndCriterion,
    )):
        target_dataset_definition = deepcopy(target_dataset.fields[set_dimension.alias].definition)

        if not isinstance(old_definition.term, (terms.ValueWrapper, terms.Function)):
            old_definition.term = target_dataset_definition

    if target_dataset and not is_metric and isinstance(old_definition, terms.BasicCriterion):
        target_dataset_definition = deepcopy(target_dataset.fields[set_dimension.alias].definition)

        if not isinstance(old_definition.left, (terms.ValueWrapper, terms.Function)):
            old_definition.left = target_dataset_definition

        if not isinstance(old_definition.right, (terms.ValueWrapper, terms.Function)):
            old_definition.right = target_dataset_definition

    set_term = set_filter.set_label
    complement_term = set_filter.complement_label

    if not set_term and not complement_term:
        set_term = "set({})".format(old_definition_sql)
        complement_term = "complement({})".format(old_definition_sql)

    if not set_filter.will_group_complement:
        complement_term = set_filter.filter.field.definition

    if is_metric or not set_filter.will_replace_referenced_dimension:
        # When keeping a referenced dimension, we name the set dimension with a custom alias, so as to have no
        # alias clashes. That prevents issues with rollups/share dimensions, given the original dimension
        # is maintained. Also, metrics need to have the same treatment, given that, unlike dimensions, they are
        # never replaced.
        set_dimension.alias = alias_selector("set({})".format(old_definition_sql))

    set_dimension.data_type = DataType.text
    set_dimension.label = (
        "Set({})".format(old_definition_sql)
        if not set_filter.set_label
        else set_filter.set_label
    )
    set_dimension.definition = (
        Case().when(old_definition, set_term).else_(complement_term)
    )

    # Necessary for set feature to work properly with data blending's field mapping.
    set_dimension.is_artificial = True

    return set_dimension


def _unwrap_field(field: Field) -> Field:
    if isinstance(field, DimensionModifier):
        return field.dimension

    return field


def _replace_field_if_needed(field: Field, fields_per_set_filter, target_dataset: 'DataSet') -> Tuple[Field]:
    set_filter = fields_per_set_filter.get(_unwrap_field(field))

    if not set_filter:
        return (field,)

    set_dimension = _make_set_dimension(set_filter, target_dataset)

    if isinstance(field, DimensionModifier):
        modified_set_dimension = deepcopy(field)
        modified_set_dimension.dimension = set_dimension
        set_dimension = modified_set_dimension

    if set_filter.will_replace_referenced_dimension and not _is_metric(set_dimension):
        # Metrics should not be replaced.
        return (set_dimension,)
    else:
        return (set_dimension, field)


def apply_set_dimensions(fields, filters, target_dataset: 'DataSet') -> List[Field]:
    """
    A transformed list of Field instances, in case `ResultSet` instances are present among the provided filters.

    :param fields: A list of Field instances.
    :param filters: A list of Filter instances.
    :param target_dataset: A DataSet instance.
    :return: A list of Field instances.
    """
    set_filters = [fltr for fltr in filters if isinstance(fltr, ResultSet)]

    if not set_filters:
        return [*fields]

    fields_per_set_filter = {
        set_filter.field: set_filter for set_filter in set_filters
    }
    fields_that_are_not_selected = set(fields_per_set_filter.keys())

    fields_with_set_dimensions = []
    for dimension_or_metric in fields:
        fields_with_set_dimensions.append(
            _replace_field_if_needed(dimension_or_metric, fields_per_set_filter, target_dataset)
        )
        unwrapped_field = _unwrap_field(dimension_or_metric)
        fields_that_are_not_selected.discard(unwrapped_field)
    fields_with_set_dimensions = flatten(fields_with_set_dimensions)

    for dimension_or_metric in fields_that_are_not_selected:
        set_filter = fields_per_set_filter[dimension_or_metric]

        if target_dataset and set_filter.field.alias not in target_dataset.fields:
            continue

        set_dimension = _make_set_dimension(set_filter, target_dataset)
        fields_with_set_dimensions.append(set_dimension)

    return fields_with_set_dimensions


def omit_set_filters(filters):
    """
    Returns all filters but the ones that are `ResultSet` instances.

    :param filters: A list of Filter instances.
    :return: A list of Filter instances.
    """
    return [fltr for fltr in filters if not isinstance(fltr, ResultSet)]



from pypika import (
    Parameter, PostgreSQLQuery,
    Table,
    functions as fn,
    terms,
)

from .base import Database


class DateTrunc(terms.Function):
    """
    Wrapper for the PostgreSQL date_trunc function
    """

    def __init__(self, field, date_format, alias=None):
        super(DateTrunc, self).__init__('DATE_TRUNC', date_format, field, alias=alias)


class PostgreSQLDatabase(Database):
    """
    PostgreSQL client that uses the psycopg module.
    """

    # The pypika query class to use for constructing queries
    query_cls = PostgreSQLQuery

    def __init__(
        self,
        host="localhost",
        port=5432,
        database=None,
        user=None,
        password=None,
        **kwargs
    ):
        super().__init__(host, port, database, **kwargs)
        self.user = user
        self.password = password

    def connect(self):
        import psycopg2

        return psycopg2.connect(
            host=self.host,
            port=self.port,
            dbname=self.database,
            user=self.user,
            password=self.password,
        )

    def trunc_date(self, field, interval):
        return DateTrunc(field, str(interval))

    def date_add(self, field, date_part, interval):
        return fn.DateAdd(str(date_part), interval, field)

    def get_column_definitions(self, schema, table, connection=None):
        columns = Table("columns", schema="INFORMATION_SCHEMA")

        columns_query = (
            PostgreSQLQuery.from_(columns, immutable=False)
            .select(columns.column_name, columns.data_type)
            .where(columns.table_schema == Parameter('%(schema)s'))
            .where(columns.field("table_name") == Parameter('%(table)s'))
            .distinct()
            .orderby(columns.column_name)
        )

        return self.fetch(str(columns_query), parameters=dict(schema=schema, table=table), connection=connection)



from fireant.dataset.totals import Rollup
from .finders import find_filters_for_totals


def adapt_for_totals_query(totals_dimension, dimensions, filters):
    """
    Adapt filters for totals query. This function will select filters for total dimensions depending on the
    apply_filter_to_totals values for the filters. A total dimension with value None indicates the base query for
    which all filters will be applied by default.

    :param totals_dimension:
    :param dimensions:
    :param filters:
    :return:
    """
    is_totals_query = totals_dimension is not None
    raw_dimensions = [dimension.dimension
                      if isinstance(dimension, Rollup)
                      else dimension
                      for dimension in dimensions]

    if not is_totals_query:
        return raw_dimensions, filters

    # Get an index to split the dimensions before and after the totals dimension
    index = [i
             for i, dimension in enumerate(dimensions)
             if dimension is totals_dimension][0]
    totals_dims = [Rollup(dimension)
                   if i >= index
                   else dimension
                   for i, dimension in enumerate(raw_dimensions)]
    totals_filters = find_filters_for_totals(filters)

    return totals_dims, totals_filters



import itertools

from fireant.queries.builder import (
    DataSetQueryBuilder,
    DimensionChoicesQueryBuilder,
    DimensionLatestQueryBuilder,
)
from fireant.utils import (
    deepcopy,
    immutable,
)


class _Container(object):
    """
    This is a list of dataset elements, metrics or dimensions, used for accessing an element by key with a dot syntax.

    Example:

    .. code-block:: python

        dataset = DataSet(
            dimensions=[
                Dimension(key='my_dimension1')
            ]
        )
        dataset.dimensions.my_dimension1
    """
    def __new__(cls, *args, **kwargs):
        """
        Because we override __getattr__ we need to implement __new__ in order for pickling to work correctly:
        https://docs.python.org/3/library/pickle.html#object.__getstate__
        """
        new_instance = super().__new__(cls)
        new_instance._items = {}
        return new_instance

    def __init__(self, items=(), key_attribute="alias"):
        self._key_attribute = key_attribute
        for item in items:
            self.add(item)

    def __deepcopy__(self, memodict={}):
        for field in self:
            memodict[id(field)] = field
        return deepcopy(self, memodict)

    def __iter__(self):
        return iter(self._items.values())

    def __getitem__(self, item):
        return getattr(self, item)

    def __getattr__(self, key):
        result = self._items.get(key)
        if result is not None:
            return result
        raise AttributeError

    def __contains__(self, item):
        from .fields import Field

        if isinstance(item, str):
            return item in self._items

        if not isinstance(item, Field):
            return False

        key = getattr(item, self._key_attribute)
        self_item = self._items.get(key)
        return item is self_item

    def __hash__(self):
        return hash((item for item in self._items.values()))

    def __eq__(self, other):
        """
        Checks if the other object is an instance of _Container and has the same number of items with matching keys.
        """
        return isinstance(other, _Container) and all(
            [
                a is not None and b is not None and a.alias == b.alias
                for a, b in itertools.zip_longest(
                    self._items.values(), getattr(other, "_items", {}).values()
                )
            ]
        )

    def add(self, item):
        key = getattr(item, self._key_attribute)
        if key in self._items:
            raise ValueError(f"Item with key {key} already exists.")

        # A very effective way to not allow reserved words but maybe not too great for performance
        if key in dir(self):
            raise ValueError(f"Reserved name {key} can not be used.")

        self._items[key] = item


class DataSet:
    """
    The DataSet class abstracts the query generation, given the fields and what not that were provided, and the
    fetching of the aforementioned query's data.
    """

    class Fields(_Container):
        pass

    def __init__(
        self,
        table,
        database,
        joins=(),
        annotation=None,
        fields=(),
        always_query_all_metrics: bool = False,
        return_additional_metadata: bool = False,
    ):
        """
        Constructor for a dataset.  Contains all the fields to initialize the dataset.

        :param table: (Required)
            A pypika Table reference. The primary table that this dataset will retrieve data from.
        :param database:  (Required)
            A Database reference. Holds the connection details used by this dataset to execute queries.
        :param fields: (Required: At least one)
            A list of fields mapping definitions of data in the data set. Fields are similar to a column in a database
            query result set. They are the values.
        :param joins:  (Optional)
            A list of join descriptions for joining additional tables.  Joined tables are only used when querying a
            metric or dimension which requires it.
        :param annotation:  (Optional)
            Annotation for fetching additional data for a dataset.
        :param always_query_all_metrics: (Default: False)
            When true, all metrics will be included in database queries in order to increase cache hits.
        :param return_additional_metadata: (Default: False)
            When true, widget data will be enveloped so extra metadata can be added to the response
            as follows: {'data': <widget data>, 'metadata': {...}}
        """
        self.table = table
        self.database = database
        self.joins = list(joins)
        self.annotation = annotation

        self.fields = DataSet.Fields(fields)

        # add query builder entry points
        self.query = DataSetQueryBuilder(self)
        self.latest = DimensionLatestQueryBuilder(self)
        self.always_query_all_metrics = always_query_all_metrics
        self.return_additional_metadata = return_additional_metadata

        for field in fields:
            if not field.definition.is_aggregate:
                field.choices = DimensionChoicesQueryBuilder(self, field)

    def __eq__(self, other):
        return isinstance(other, DataSet) and self.fields == other.fields

    def __repr__(self):
        return "DataSet(fields=[{}])".format(",".join([repr(f) for f in self.fields]))

    def __hash__(self):
        return hash(
            (
                self.table,
                self.database.database,
                tuple(self.joins),
                self.fields,
                self.always_query_all_metrics,
            )
        )

    @immutable
    def extra_fields(self, *fields):
        for field in fields:
            self.fields.add(field)

    def blend(self, other):
        """
        Returns a Data Set blender which enables to execute queries on multiple data sets and combine them.
        """
        from .data_blending import DataSetBlenderBuilder

        return DataSetBlenderBuilder(self, other)



from typing import Dict, Iterable, List, TYPE_CHECKING, Type, Union

from fireant.dataset.fields import DataType
from fireant.dataset.intervals import DatetimeInterval
from fireant.dataset.totals import scrub_totals_from_share_results
from fireant.reference_helpers import (
    apply_reference_filters,
    reference_alias,
)
from fireant.utils import (
    alias_selector,
    immutable,
)
from .query_builder import (
    QueryBuilder,
    QueryException,
    ReferenceQueryBuilderMixin,
    WidgetQueryBuilderMixin,
    add_hints,
)
from .. import special_cases
from ..execution import fetch_data
from ..finders import (
    find_and_group_references_for_dimensions,
    find_field_in_modified_field,
    find_metrics_for_widgets,
    find_operations_for_widgets,
    find_share_dimensions,
)
from ..pagination import paginate
from ..sql_transformer import (
    make_slicer_query,
    make_slicer_query_with_totals_and_references,
)

if TYPE_CHECKING:
    from pypika import PyPikaQueryBuilder


class DataSetQueryBuilder(
    ReferenceQueryBuilderMixin, WidgetQueryBuilderMixin, QueryBuilder
):
    """
    Data Set queries consist of widgets, dimensions, filters, orders by and references. At least one or more widgets
    is required. All others are optional.
    """

    def __init__(self, dataset):
        super().__init__(dataset)
        self._totals_dimensions = set()
        self._apply_filter_to_totals = []

    def __call__(self, *args, **kwargs):
        return self

    @immutable
    def filter(self, *filters, apply_to_totals=True):
        """
        Add one or more filters when building a dataset query.

        :param filters:
            Filters to add to the query
        :param apply_to_totals:
            Whether filters should apply to totals or not
        :return:
            A copy of the query with the filters added.
        """
        self._filters += [f for f in filters]
        self._apply_filter_to_totals += [apply_to_totals] * len(filters)

    @property
    def reference_groups(self):
        return list(
            find_and_group_references_for_dimensions(
                self.dimensions, self._references
            ).values()
        )

    @property
    def sql(self) -> List[Type['PyPikaQueryBuilder']]:
        """
        Serialize this query builder to a list of Pypika/SQL queries. This function will return one query for every
        combination of reference and rolled up dimension (including null options).

        This collects all of the metrics in each widget, dimensions, and filters and builds a corresponding pypika query
        to fetch the data.  When references are used, the base query normally produced is wrapped in an outer query and
        a query for each reference is joined based on the referenced dimension shifted.

        :return: a list of Pypika's Query subclass instances.
        """
        # First run validation for the query on all widgets
        self._validate()

        dimensions = self.dimensions

        metrics = find_metrics_for_widgets(self._widgets)
        operations = find_operations_for_widgets(self._widgets)
        share_dimensions = find_share_dimensions(dimensions, operations)

        queries = make_slicer_query_with_totals_and_references(
            database=self.dataset.database,
            table=self.table,
            joins=self.dataset.joins,
            dimensions=dimensions,
            metrics=metrics,
            operations=operations,
            filters=self.filters,
            references=self._references,
            orders=self.orders,
            share_dimensions=share_dimensions,
        )

        return [self._apply_pagination(query) for query in queries]

    def fetch(self, hint=None) -> Union[Iterable[Dict], Dict]:
        """
        Fetch the data for this query and transform it into the widgets.

        :param hint:
            A query hint label used with database vendors which support it. Adds a label comment to the query.
        :return:
            A list of dict (JSON) objects containing the widget configurations.
        """
        queries = add_hints(self.sql, hint)

        operations = find_operations_for_widgets(self._widgets)
        dimensions = self.dimensions

        share_dimensions = find_share_dimensions(dimensions, operations)

        annotation_frame = None
        if dimensions and self.dataset.annotation:
            alignment_dimension_alias = (
                self.dataset.annotation.dataset_alignment_field_alias
            )
            first_dimension = find_field_in_modified_field(dimensions[0])

            if first_dimension.alias == alignment_dimension_alias:
                annotation_frame = self.fetch_annotation()

        max_rows_returned, data_frame = fetch_data(
            self.dataset.database,
            queries,
            dimensions,
            share_dimensions,
            self.reference_groups,
        )

        # Apply reference filters
        for reference in self._references:
            data_frame = apply_reference_filters(data_frame, reference)

        # Apply operations
        for operation in operations:
            for reference in [None] + self._references:
                df_key = alias_selector(reference_alias(operation, reference))
                data_frame[df_key] = operation.apply(data_frame, reference)

        data_frame = scrub_totals_from_share_results(data_frame, dimensions)
        data_frame = special_cases.apply_operations_to_data_frame(
            operations, data_frame
        )

        data_frame = paginate(
            data_frame,
            self._widgets,
            orders=self.orders,
            limit=self._client_limit,
            offset=self._client_offset,
        )

        # Apply transformations
        widget_data = [
            widget.transform(
                data_frame,
                dimensions,
                self._references,
                annotation_frame,
            )
            for widget in self._widgets
        ]

        return self._transform_for_return(widget_data, max_rows_returned=max_rows_returned)

    def fetch_annotation(self):
        """
        Fetch annotation data for this query builder.

        :return:
            A data frame containing the annotation data.
        """
        annotation = self.dataset.annotation

        # Fetch filters for the dataset's alignment dimension from this query builder
        dataset_alignment_dimension_filters = self.fetch_query_filters(
            annotation.dataset_alignment_field_alias
        )

        # Update fields in filters for the dataset's alignment dimension to the annotation's alignment field
        annotation_alignment_dimension_filters = [
            dataset_alignment_filter.for_(annotation.alignment_field)
            for dataset_alignment_filter in dataset_alignment_dimension_filters
        ]

        annotation_alignment_field = annotation.alignment_field
        if annotation_alignment_field.data_type == DataType.date:
            dataset_alignment_dimension = self.fetch_query_dimension(
                annotation.dataset_alignment_field_alias
            )

            if hasattr(dataset_alignment_dimension, "interval_key"):
                # Use the interval key of the dataset's alignment dimension for the annotation's alignment field
                # Otherwise we would need to copy it to prevent issues from patching directly
                annotation_alignment_field = DatetimeInterval(
                    annotation.alignment_field, dataset_alignment_dimension.interval_key
                )

        annotation_dimensions = [annotation_alignment_field, annotation.field]

        annotation_query = make_slicer_query(
            database=self.dataset.database,
            base_table=annotation.table,
            dimensions=annotation_dimensions,
            filters=annotation_alignment_dimension_filters,
        )

        _, annotation_df = fetch_data(
            self.dataset.database, [annotation_query], [annotation.alignment_field]
        )

        return annotation_df

    def fetch_query_filters(self, dimension_alias):
        """
        Fetch all filters matching the given dimension alias from this query builder. All fields of a filter
        (e.g. complex fields) have to match the alias and use the base table of the dataset.

        :param dimension_alias:
            An alias of a dimension.
        :return:
            A list of filters for the given dimension alias.
        """
        dimension_filters = []

        for filter_ in self.filters:
            filter_fields = [
                field
                for field in filter_.definition.fields_()
                if all(table == self.dataset.table for table in field.tables_)
            ]

            if not filter_fields:
                continue

            if all(field.name == dimension_alias for field in filter_fields):
                dimension_filters.append(filter_)

        return dimension_filters

    def fetch_query_dimension(self, dimension_alias):
        """
        Fetch a dimension matching the given dimension alias from this query builder.

        :param dimension_alias:
            An alias of a dimension.
        :return:
            A dimension (Field) of this query builder matching the dimension alias.
        """
        for dimension in self.dimensions:
            unwrapped_dimension = find_field_in_modified_field(dimension)
            if unwrapped_dimension.alias == dimension_alias:
                return dimension

        return None

    def plot(self):
        try:
            from IPython.display import display
        except ImportError:
            raise QueryException(
                "Optional dependency ipython missing. Please install fireant[ipython] to use plot."
            )

        widgets = self.fetch()
        for widget in reversed(widgets):
            display(widget)

    def __str__(self):
        return str(self.sql)

    def __repr__(self):
        return ".".join(
            [
                "dataset",
                "query",
                *["widget({})".format(repr(widget)) for widget in self._widgets],
                *[
                    "dimension({})".format(repr(dimension))
                    for dimension in self._dimensions
                ],
                *[
                    "filter({}{})".format(
                        repr(f),
                        ", apply_filter_to_totals=True"
                        if apply_filter_to_totals
                        else "",
                    )
                    for f, apply_filter_to_totals in zip(
                        self._filters, self._apply_filter_to_totals
                    )
                ],
                *[
                    "reference({})".format(repr(reference))
                    for reference in self._references
                ],
                *[
                    "orderby({}, {})".format(definition.alias, orientation)
                    for (definition, orientation) in self.orders
                ],
            ]
        )



from fireant.dataset.intervals import DatetimeInterval
from fireant.utils import alias_selector


def make_term_for_field(field, window=None):
    """
    Makes a list of pypika terms for a given dataset field.

    :param field:
        A field from a dataset.
    :param window:
        A window function to apply to the dimension definition if it is a continuous dimension.
    :return:
        a list of terms required to select and group by in a SQL query given a dataset dimension. This list will contain
        either one or two elements. A second element will be included if the dimension has a definition for its display
        field.
    """
    f_alias = alias_selector(field.alias)

    if window and isinstance(field, DatetimeInterval):
        return window(field.definition, field.interval_key).as_(f_alias)

    return field.definition.as_(f_alias)



from _csv import QUOTE_MINIMAL
from typing import Iterable

from fireant.dataset.fields import Field
from .pandas import Pandas


class CSV(Pandas):
    def __init__(
        self, metric: Field, *metrics: Iterable[Field], group_pagination=False, **kwargs
    ):
        super().__init__(metric, *metrics, **kwargs)
        self.group_pagination = group_pagination

    def transform(
        self,
        data_frame,
        dimensions,
        references,
        annotation_frame=None,
        use_raw_values=None,
    ):
        result_df = super(CSV, self).transform(
            data_frame, dimensions, references, use_raw_values=True
        )
        # Unset the column level names because they're a bit confusing in a csv file
        result_df.columns.names = [None] * len(result_df.columns.names)
        return result_df.to_csv(na_rep="", quoting=QUOTE_MINIMAL)



import numpy as np
import pandas as pd

from fireant.dataset.references import calculate_delta_percent
from fireant.dataset.totals import get_totals_marker_for_dtype
from fireant.utils import (
    alias_selector,
    reduce_data_frame_levels,
)
from .fields import (
    DataType,
    Field,
)
from ..reference_helpers import reference_alias, reference_type_alias


def _extract_key_or_arg(data_frame, key):
    return data_frame[key] if key in data_frame else key


class Operation:
    """
    The `Operation` class represents an operation in the `DataSet` API.
    """

    def apply(self, data_frame, reference):
        raise NotImplementedError()

    @property
    def metrics(self):
        raise NotImplementedError()

    @property
    def operations(self):
        return []


class _BaseOperation(Operation):
    data_type = DataType.number

    def __init__(
        self,
        alias,
        label,
        args,
        prefix: str = None,
        suffix: str = None,
        thousands: str = None,
        precision: int = None,
    ):
        self.alias = alias
        self.label = label
        self.args = args
        self.prefix = prefix
        self.suffix = suffix
        self.thousands = thousands
        self.precision = precision

    @property
    def metrics(self):
        return [metric for metric in self.args if isinstance(metric, Field)]

    @property
    def operations(self):
        return [
            op_and_children
            for operation in self.args
            if isinstance(operation, Operation)
            for op_and_children in [operation] + operation.operations
        ]

    def _group_levels(self, index):
        """
        Get the index levels that need to be grouped. This is to avoid apply the cumulative function across separate
        dimensions. Only the first dimension should be accumulated across.

        :param index:
        :return:
        """
        return index.names[1:]

    def __repr__(self):
        return self.alias


class _Cumulative(_BaseOperation):
    def __init__(self, arg):
        super(_Cumulative, self).__init__(
            alias="{}({})".format(
                self.__class__.__name__.lower(), getattr(arg, "alias", arg)
            ),
            label="{}({})".format(self.__class__.__name__, getattr(arg, "label", arg)),
            args=[arg],
            prefix=getattr(arg, "prefix"),
            suffix=getattr(arg, "suffix"),
            thousands=getattr(arg, "thousands"),
            precision=getattr(arg, "precision"),
        )

    @property
    def metric(self):
        return self.args[0]

    def apply(self, data_frame, reference):
        if reference and reference.delta_percent:
            return self._apply_cumulative_for_reference_delta_percent(data_frame, reference)

        f_metric_alias = alias_selector(reference_alias(self.metric, reference))
        return self._apply_cumulative(data_frame, f_metric_alias)

    def _apply_cumulative(self, data_frame, reference):
        raise NotImplementedError()

    def _apply_cumulative_for_reference_delta_percent(self, data_frame, reference):
        """
        When a delta percent reference is combined with a cumulative operation, the delta percent values need to be
        calculated based on the result of performing the operation on both the base values as well as
        the reference values. The correct result can not be obtained by simply applying the operation to the delta
        percent values.
        """
        # apply the operation on the original reference values
        original_reference_alias = alias_selector(reference_type_alias(self.metric, reference))
        reference_values_after_operation = self._apply_cumulative(data_frame, original_reference_alias)

        # get the base values on which the operation is already performed
        base_values_after_operation_key = alias_selector(self.alias)
        base_values_after_operation = data_frame[base_values_after_operation_key]

        # recalculate the delta using the values on which the operation is already performed
        ref_delta_df = base_values_after_operation.subtract(
            reference_values_after_operation, fill_value=0
        )
        # recalculate the delta percent
        return calculate_delta_percent(reference_values_after_operation, ref_delta_df)


class CumSum(_Cumulative):
    def _apply_cumulative(self, data_frame, f_metric_alias):
        if isinstance(data_frame.index, pd.MultiIndex) and not data_frame.empty:
            levels = self._group_levels(data_frame.index)

            return data_frame[f_metric_alias].groupby(level=levels).cumsum()

        return data_frame[f_metric_alias].cumsum()


class CumProd(_Cumulative):
    def _apply_cumulative(self, data_frame, f_metric_alias):
        if isinstance(data_frame.index, pd.MultiIndex) and not data_frame.empty:
            levels = self._group_levels(data_frame.index)

            return data_frame[f_metric_alias].groupby(level=levels).cumprod()

        return data_frame[f_metric_alias].cumprod()


class CumMean(_Cumulative):
    @staticmethod
    def cummean(x):
        return x.cumsum() / np.arange(1, len(x) + 1)

    def _apply_cumulative(self, data_frame, f_metric_alias):
        if isinstance(data_frame.index, pd.MultiIndex) and not data_frame.empty:
            levels = self._group_levels(data_frame.index)

            return data_frame[f_metric_alias].groupby(level=levels).apply(self.cummean)

        return self.cummean(data_frame[f_metric_alias])


class RollingOperation(_BaseOperation):
    def __init__(self, arg, window, min_periods=None):
        super(RollingOperation, self).__init__(
            alias="{}({},{})".format(
                self.__class__.__name__.lower(), getattr(arg, "alias", arg), window
            ),
            label="{}({},{})".format(
                self.__class__.__name__, getattr(arg, "label", arg), window
            ),
            args=[arg],
            prefix=getattr(arg, "prefix"),
            suffix=getattr(arg, "suffix"),
            thousands=getattr(arg, "thousands"),
            precision=getattr(arg, "precision"),
        )
        self.window = window
        self.min_periods = min_periods

    def _should_adjust(self, other_operations):
        # Need to figure out if this rolling operation is has the largest window, and if it's the first of multiple
        # rolling operations if there are more than one operation sharing the largest window.
        first_max_rolling = list(
            sorted(other_operations, key=lambda operation: operation.window)
        )[0]

        return first_max_rolling is self

    def apply(self, data_frame, reference):
        raise NotImplementedError()


class RollingMean(RollingOperation):
    def rolling_mean(self, x):
        return x.rolling(self.window, self.min_periods).mean()

    def apply(self, data_frame, reference):
        (arg,) = self.args
        df_alias = alias_selector(reference_alias(arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_alias].groupby(level=levels).apply(self.rolling_mean)

        return self.rolling_mean(data_frame[df_alias])


class Share(_BaseOperation):
    def __init__(self, metric: Field, over: Field = None, precision=2):
        super(Share, self).__init__(
            alias="share({},{})".format(
                getattr(metric, "alias", metric), getattr(over, "alias", over),
            ),
            label="Share of {} over {}".format(
                getattr(metric, "label", metric), getattr(over, "label", over)
            ),
            args=[metric, over],
            suffix="%",
            precision=precision,
        )

    @property
    def metrics(self):
        return [self.args[0]]

    @property
    def metric(self):
        return self.args[0]

    @property
    def over(self):
        return self.args[1]

    def _get_metric_alias(self, reference):
        return alias_selector(reference_alias(self.metric, reference))

    def apply(self, data_frame, reference):
        if reference and reference.delta:
            return self._apply_share_for_reference_delta(data_frame, reference)

        f_metric_alias = alias_selector(reference_alias(self.metric, reference))
        return self._apply_share(data_frame, f_metric_alias)

    def _apply_share_for_reference_delta(self, data_frame, reference):
        # apply the operation on the original reference values
        original_reference_alias = alias_selector(reference_type_alias(self.metric, reference))
        reference_values_after_operation = self._apply_share(data_frame, original_reference_alias)

        # get the base values on which the operation is already performed
        base_values_after_operation_key = alias_selector(self.alias)
        base_values_after_operation = data_frame[base_values_after_operation_key]

        # recalculate the delta using the values on which the operation is already performed
        ref_delta_df = base_values_after_operation.subtract(
            reference_values_after_operation, fill_value=0
        )
        if reference.delta_percent:
            # recalculate the delta percent
            ref_delta_df = calculate_delta_percent(reference_values_after_operation, ref_delta_df)

        return ref_delta_df

    def _apply_share(self, data_frame, f_metric_alias):
        if self.over is None:
            df = data_frame[f_metric_alias]
            return 100 * df / df

        if not isinstance(data_frame.index, pd.MultiIndex):
            marker = get_totals_marker_for_dtype(data_frame.index.dtype)
            totals = data_frame.loc[marker, f_metric_alias]
            if totals == 0:
                return np.nan
            return 100 * data_frame[f_metric_alias] / totals

        f_over_alias = alias_selector(self.over.alias)
        idx = data_frame.index.names.index(f_over_alias)
        group_levels = data_frame.index.names[idx:]
        over_dim_value = get_totals_marker_for_dtype(data_frame.index.levels[idx].dtype)
        totals_alias = (slice(None),) * idx + (slice(over_dim_value, over_dim_value),)

        totals = reduce_data_frame_levels(
            data_frame.loc[totals_alias, f_metric_alias], group_levels
        )

        def apply_totals(group_df):
            if not isinstance(totals, pd.Series):
                return 100 * group_df / totals

            n_index_levels = len(totals.index.names)
            extra_level_names = group_df.index.names[n_index_levels:]
            group_df = group_df.reset_index(extra_level_names, drop=True)
            share = 100 * group_df / totals[group_df.index]
            return pd.Series(share.values, index=group_df.index)

        return (
            data_frame[f_metric_alias]
            .groupby(level=group_levels)
            .apply(apply_totals)
            .reorder_levels(order=data_frame.index.names)
            .sort_index()
        )



from typing import Union

import pandas as pd

from fireant.dataset.fields import Field
from fireant.dataset.operations import Operation
from fireant.dataset.references import Reference
from fireant.exceptions import DataSetException
from fireant.reference_helpers import (
    reference_alias,
    reference_label,
    reference_prefix,
    reference_suffix,
)
from fireant.utils import (
    alias_selector,
    deepcopy,
    immutable,
)


class MetricRequiredException(DataSetException):
    pass


class Widget:
    def __init__(self, *items: Union[Field, Operation]):
        self.items = list(items)

    @immutable
    def item(self, item):
        self.items.append(item)

    @property
    def metrics(self):
        if 0 == len(self.items):
            raise MetricRequiredException(str(self))

        return [
            metric
            for group in self.items
            for metric in getattr(group, "metrics", [group])
        ]

    def __deepcopy__(self, memodict={}):
        for item in self.items:
            memodict[id(item)] = item
        return deepcopy(self, memodict)

    @property
    def operations(self):
        return [item for item in self.items if isinstance(item, Operation)]

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.items == other.items

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__, ",".join(str(m) for m in self.items)
        )


class TransformableWidget(Widget):
    # This attribute can be overridden in order to sort groups. Useful in cases like for charts where sorting
    # should be applied to the number of series rather than the number of data points.
    group_pagination = False

    @staticmethod
    def hide_data_frame_indexes(data_frame, dimensions_to_hide):
        data_frame_indexes = (
            [data_frame.index.name] if not isinstance(data_frame.index, pd.MultiIndex) else data_frame.index.names
        )

        for dimension in dimensions_to_hide:
            dimesion_alias = alias_selector(dimension.alias)
            if dimesion_alias in data_frame_indexes:
                data_frame.reset_index(level=dimesion_alias, drop=True, inplace=True)

    def transform(self, data_frame, dimensions, references, annotation_frame=None):
        """
        - Main entry point -

        Transformers the result set `pd.DataFrame` from a dataset query into the output format for this specific widget
        type.

        :param data_frame:
            The data frame containing the data. Index must match the dimensions parameter.
        :param dimensions:
            A list of dimensions that are being rendered.
        :param references:
            A list of references that are being rendered.
        :param annotation_frame:
            A data frame containing the annotation data.
        :return:
            A dict meant to be dumped as JSON.
        """
        raise NotImplementedError()


class ReferenceItem:
    def __init__(self, item, reference):
        assert isinstance(reference, Reference)
        self.data_type = item.data_type
        self.alias = reference_alias(item, reference)
        self.label = reference_label(item, reference)
        self.prefix = reference_prefix(item, reference)
        self.suffix = reference_suffix(item, reference)
        self.thousands = item.thousands
        self.precision = item.precision



from collections import Iterable
from enum import Enum
from functools import wraps
from numbers import Number
from typing import Type, Union

from pypika import Field as PyPikaField
from pypika.enums import Arithmetic
from pypika.terms import (
    ArithmeticExpression,
    Function,
    Mod,
    Node,
    Pow,
)

from .filters import (
    AntiPatternFilter,
    BooleanFilter,
    ComparatorFilter,
    ComparisonOperator,
    ContainsFilter,
    ExcludesFilter,
    PatternFilter,
    RangeFilter,
    VoidFilter,
)
from ..utils import immutable


class DataType(Enum):
    date = 1
    text = 2
    number = 3
    boolean = 4

    def __repr__(self):
        return self.name


CONTINUOUS_TYPES = [DataType.number, DataType.date]
DISCRETE_TYPES = [DataType.text, DataType.boolean]


class DataSetFilterException(Exception):
    def __init__(self, msg, allowed):
        super().__init__(msg)
        self.allowed = allowed


def restrict_types(allowed):
    def wrapper(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            if self.data_type not in allowed:
                msg = "`{func}` filter can only be used on fields with the following data types: {types}.".format(
                    func=func.__name__, types=", ".join(map(repr, allowed))
                )
                raise DataSetFilterException(msg, allowed)
            return func(self, *args, **kwargs)

        return inner

    return wrapper


class Field(Node):
    """
    The `Field` class represents a field in the `DataSet` object. A field is a mapping to a column in a database query.

    :param alias:
        A unique identifier used to identify the metric when writing dataset queries. This value must be unique over
        the metrics in the dataset.

    :param definition:
        A pypika expression which is used to select the value when building SQL queries. For metrics, this query
        **must** be aggregated, since queries always use a ``GROUP BY`` clause an metrics are not used as a group.

    :param data_type: {Number, Text, Boolean, Date}
        When True, the field's definition should be treated as an aggregate expression.

    :param label: (optional)
        A display value used for the metric. This is used for rendering the labels within the visualizations. If not
        set, the alias will be used as the default.

    :param hint_table: (optional)
        An optional table for fetching field choices. If this table is not set, the base table from the field
        definition will be used.

    :param precision: (optional)
        A precision value for rounding decimals. By default, no rounding will be applied.

    :param prefix: (optional)
        A prefix for rendering labels in visualizations such as '$'.

    :param suffix: (optional)
        A suffix for rendering labels in visualizations such as '€'.

    :param thousands: (optional)
        A character to be used as a thousand separator when rendering values.

    :param hyperlink_template: (optional)
        A url string that can also reference other fields in a dataset. To reference a field put its alias
        between {}. This is only used in certain widgets (i.e. react table).

    :param fetch_only: (optional)
        Whether the field data should be ignored in widgets. This is useful for not displaying hyperlink
        dependencies, which might be necessary only for the purpose of generating a hyperlink and have no
        effect on the dimension grouping.
    """

    def __init__(
        self,
        alias: str,
        definition: Union[PyPikaField, Type[Function]],
        data_type: DataType = DataType.number,
        label=None,
        hint_table=None,
        prefix: str = None,
        suffix: str = None,
        thousands: str = None,
        precision: int = None,
        hyperlink_template: str = None,
        fetch_only: bool = False,
    ):
        self.alias = alias
        self.data_type = data_type
        self.definition = definition
        self.label = label if label is not None else alias
        self.hint_table = hint_table
        self.prefix = prefix
        self.suffix = suffix
        self.thousands = thousands
        self.precision = precision
        self.hyperlink_template = hyperlink_template
        self.fetch_only = fetch_only

        # An artificial field is created dynamically as the query is mounted, instead of being defined during
        # instantiation. That's the case for set dimensions, for instance. The only practical aspect of this
        # is that artificial dimensions are hashed differently (see dunder hash method). Instead of using the
        # memory reference, artificial dimensions use a set of its fields.
        self.is_artificial = False

    @property
    def is_aggregate(self):
        return self.definition.is_aggregate

    @property
    def is_wrapped(self) -> bool:
        """
        This allows calling code can easily tell whether the field has been wrapped.
        This occurs if data blending is being used.
        """
        return bool(getattr(self.definition, 'definition', False))

    def eq(self, other: Number) -> ComparatorFilter:
        return ComparatorFilter(self, ComparisonOperator.eq, other)

    def ne(self, other: Number) -> ComparatorFilter:
        return ComparatorFilter(self, ComparisonOperator.ne, other)

    @restrict_types(CONTINUOUS_TYPES)
    def gt(self, other: Number) -> ComparatorFilter:
        return ComparatorFilter(self, ComparisonOperator.gt, other)

    @restrict_types(CONTINUOUS_TYPES)
    def ge(self, other: Number) -> ComparatorFilter:
        return ComparatorFilter(self, ComparisonOperator.gte, other)

    @restrict_types(CONTINUOUS_TYPES)
    def lt(self, other: Number):
        return ComparatorFilter(self, ComparisonOperator.lt, other)

    @restrict_types(CONTINUOUS_TYPES)
    def le(self, other: Number) -> ComparatorFilter:
        return ComparatorFilter(self, ComparisonOperator.lte, other)

    @restrict_types(CONTINUOUS_TYPES)
    def between(self, lower, upper) -> RangeFilter:
        """
        Creates a filter to filter a dataset query.

        :param lower:
            The start time of the filter. This is the beginning of the window for which results should be included.
        :param upper:
            The stop time of the filter. This is the end of the window for which results should be included.
        :return:
            A dataset query filter used to filter a dataset query to results where this dimension is between the values
            start and stop.
        """
        return RangeFilter(self, lower, upper)

    def isin(self, values: Iterable) -> ContainsFilter:
        """
        Creates a filter that filters to rows where

        :param values:
            An iterable of value to constrain the dataset query results by.

        :return:
            A dataset query filter used to filter a dataset query to results where this dimension is one of a set of
            values. Opposite of #notin.
        """
        return ContainsFilter(self, values)

    def notin(self, values: Iterable) -> ExcludesFilter:
        """
        Creates a filter to filter a dataset query.

        :param values:
            An iterable of value to constrain the dataset query results by.

        :return:
            A dataset query filter used to filter a dataset query to results where this dimension is *not* one of a
            set of
            values. Opposite of #isin.
        """
        return ExcludesFilter(self, values)

    @restrict_types([DataType.text])
    def like(self, pattern: str, *patterns: Iterable) -> PatternFilter:
        return PatternFilter(self, pattern, *patterns)

    @restrict_types([DataType.text])
    def not_like(self, pattern: str, *patterns: Iterable) -> AntiPatternFilter:
        return AntiPatternFilter(self, pattern, *patterns)

    @restrict_types([DataType.boolean])
    def is_(self, value: bool) -> BooleanFilter:
        """
        Creates a filter to filter a dataset query.

        :param value:
            True or False
        :return:
            A dataset query filter used to filter a dataset query to results where this dimension is True or False.
        """
        return BooleanFilter(self, value)

    def void(self) -> VoidFilter:
        return VoidFilter(self)

    def __eq__(self, other):
        return self.eq(other)

    def __ne__(self, other):
        return self.ne(other)

    def __gt__(self, other):
        return self.gt(other)

    def __ge__(self, other):
        return self.ge(other)

    def __lt__(self, other):
        return self.lt(other)

    def __le__(self, other):
        return self.le(other)

    def __add__(self, other):
        return ArithmeticExpression(Arithmetic.add, self, other)

    def __sub__(self, other):
        return ArithmeticExpression(Arithmetic.sub, self, other)

    def __mul__(self, other):
        return ArithmeticExpression(Arithmetic.mul, self, other)

    def __truediv__(self, other):
        return ArithmeticExpression(Arithmetic.div, self, other)

    def __pow__(self, other):
        return Pow(self, other)

    def __mod__(self, other):
        return Mod(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        return self.__sub__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rtruediv__(self, other):
        return self.__truediv__(other)

    def __repr__(self):
        return "{alias}[{data_type}]".format(alias=self.alias, data_type=self.data_type)

    def __hash__(self):
        if not hasattr(self, "alias"):
            """
            Python's deep copy built-in are a problem when:
                Have classes that must be hashed and contain reference cycles, and
                Don't ensure hash-related (and equality related) invariants are established at object
                construction, not just initialization
            Since that's our case, we do this simple check to avoid exceptions during deep copy. Note that any
            of the variables in this class would amount to the same. That said it's best to use a non nullable
            positional argument.
            """
            return hash(None)

        if self.is_artificial:
            """
            Python's deep copy built-in are a problem when:
                Handling artificial fields (e.g. set field) in data blending.
            Given artificial fields are dynamically generated, they can't be properly compared when hashed
            by memory id. Therefore we hash them differently.
            """
            return hash((self.alias, self.label, self.data_type, self.definition))

        return id(self)

    def for_(self, replacement):
        return replacement

    def get_sql(self, *args, **kwargs) -> str:
        return self.definition.get_sql(*args, **kwargs)

    @property
    @immutable
    def share(self):
        self._share = True
        return self



from typing import List

from pypika import Order

from fireant.utils import alias_selector
from .query_builder import (
    QueryBuilder,
    add_hints,
    get_column_names,
)
from ..execution import fetch_data
from ..field_helper import make_term_for_field
from ..finders import find_joins_for_tables
from ..sql_transformer import make_slicer_query
from ...formats import display_value


class DimensionChoicesQueryBuilder(QueryBuilder):
    """
    This builder is used for building dataset queries for fetching the choices for a dimension given a set of filters.
    """

    def __init__(self, dataset, dimension):
        super().__init__(dataset)

        self.hint_table = getattr(dimension, "hint_table", None)
        self._dimensions.append(dimension)

        # TODO remove after 3.0.0
        display_alias = dimension.alias + "_display"
        if display_alias in dataset.fields:
            self._dimensions.append(dataset.fields[display_alias])

    def _extract_hint_filters(self):
        """
        Extracts filters that can be applied when using the hint table.

        :return:
            A list of filters.
        """
        base_table = self.dataset.table
        hint_column_names = get_column_names(self.dataset.database, self.hint_table)

        filters = []
        for filter_ in self.filters:
            base_fields = [
                field
                for field in filter_.definition.fields_()
                if all(table == base_table for table in field.tables_)
            ]

            join_tables = [
                table
                for field in filter_.definition.fields_()
                for table in field.tables_
                if table != base_table
            ]

            required_joins = find_joins_for_tables(
                self.dataset.joins, self.dataset.table, join_tables
            )

            base_fields.extend(
                [
                    field
                    for join in required_joins
                    for field in join.criterion.fields_()
                    if all(table == base_table for table in field.tables_)
                ]
            )

            if all(field.name in hint_column_names for field in base_fields):
                filters.append(filter_)

        return filters

    def _make_terms_for_hint_dimensions(self):
        """
        Makes a list pypika terms using the hint table instead of their original table.

        :return:
            A list of pypika terms.
        """
        dimension_terms = []
        for dimension in self.dimensions:
            dimension_term = make_term_for_field(
                dimension, self.dataset.database.trunc_date
            )
            dimension_term = dimension_term.replace_table(
                dimension_term.table, self.hint_table
            )
            dimension_terms.append(dimension_term)

        return dimension_terms

    @property
    def sql(self):
        """
        Serializes this query builder as a set of SQL queries. This method will always return a list of one query since
        only one query is required to retrieve dimension choices.

        The dataset query extends this with metrics, references, and totals.
        """
        dimensions = [] if self.hint_table else self.dimensions

        filters = self._extract_hint_filters() if self.hint_table else self.filters

        query = (
            make_slicer_query(
                database=self.dataset.database,
                base_table=self.dataset.table,
                joins=self.dataset.joins,
                dimensions=dimensions,
                filters=filters,
            )
            .limit(self._query_limit)
            .offset(self._query_offset)
        )

        if self.hint_table:
            hint_dimension_terms = self._make_terms_for_hint_dimensions()
            query = query.select(*hint_dimension_terms).groupby(*hint_dimension_terms)
            query = query.replace_table(self.dataset.table, self.hint_table)

        return [query]

    def fetch(self, hint=None, force_include=()) -> List[str]:
        """
        Fetch the data for this query and transform it into the widgets.

        :param hint:
            For database vendors that support it, add a query hint to collect analytics on the queries triggered by
            fireant.
        :param force_include:
            A list of dimension values to include in the result set. This can be used to avoid having necessary results
            cut off due to the pagination.  These results will be returned at the head of the results.
        :return:
            A list of dict (JSON) objects containing the widget configurations.
        """
        query = add_hints(self.sql, hint)[0]
        dimension = self.dimensions[0]
        alias_definition = dimension.definition.as_(alias_selector(dimension.alias))
        dimension_definition = dimension.definition

        if self.hint_table:
            alias_definition = alias_definition.replace_table(
                alias_definition.table, self.hint_table
            )
            dimension_definition = dimension.definition.replace_table(
                dimension_definition.table, self.hint_table
            )

        if force_include:
            include = self.dataset.database.to_char(dimension_definition).isin(
                [str(x) for x in force_include]
            )

            # Ensure that these values are included
            query = query.orderby(include, order=Order.desc)

        # Filter out NULL values from choices
        query = query.where(dimension_definition.notnull())

        # Order by the dimension definition that the choices are for
        query = query.orderby(alias_definition)
        max_rows_returned, data = fetch_data(self.dataset.database, [query], self.dimensions)

        if len(data.index.names) > 1:
            display_alias = data.index.names[1]
            data.reset_index(display_alias, inplace=True)
            choices = data[display_alias]

        else:
            data["display"] = data.index.tolist()
            choices = data["display"]

        dimension_display = self.dimensions[-1]
        choices = choices.map(lambda raw: display_value(raw, dimension_display) or raw)
        return self._transform_for_return(choices, max_rows_returned=max_rows_returned)

    def __repr__(self):
        return ".".join(
            ["dataset", self._dimensions[0].alias, "choices"]
            + ["filter({})".format(repr(f)) for f in self._filters]
        )



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from pypika import (
    Parameter, Table, functions as fn,
    terms,
)
from pypika.dialects import SnowflakeQuery

from .base import Database

try:
    from snowflake import connector as snowflake
except:
    pass

IGNORED_SCHEMAS = {'INFORMATION_SCHEMA'}


class Trunc(terms.Function):
    """
    Wrapper for TRUNC function for truncating dates.
    """

    def __init__(self, field, date_format, alias=None):
        super(Trunc, self).__init__('TRUNC', field, date_format, alias=alias)


class SnowflakeDatabase(Database):
    """
    Snowflake client.
    """

    # The pypika query class to use for constructing queries
    query_cls = SnowflakeQuery

    DATETIME_INTERVALS = {
        'hour': 'HH',
        'day': 'DD',
        'week': 'IW',
        'month': 'MM',
        'quarter': 'Q',
        'year': 'Y'
    }
    _private_key = None

    def __init__(self, user='snowflake', password=None,
                 account='snowflake', database='snowflake',
                 private_key_data=None, private_key_password=None,
                 region=None, warehouse=None, **kwags):
        super(SnowflakeDatabase, self).__init__(database=database, **kwags)
        self.user = user
        self.password = password
        self.account = account
        self.private_key_data = private_key_data
        self.private_key_password = private_key_password
        self.region = region
        self.warehouse = warehouse

    def connect(self):
        import snowflake

        return snowflake.connector.connect(database=self.database,
                                           account=self.account,
                                           user=self.user,
                                           password=self.password,
                                           private_key=self._get_private_key(),
                                           region=self.region,
                                           warehouse=self.warehouse)

    def trunc_date(self, field, interval):
        trunc_date_interval = self.DATETIME_INTERVALS.get(str(interval), 'DD')
        return Trunc(field, trunc_date_interval)

    def date_add(self, field, date_part, interval):
        return fn.TimestampAdd(str(date_part), interval, field)

    def _get_private_key(self):
        if self._private_key is None:
            self._private_key = self._load_private_key_data()

        return self._private_key

    def _load_private_key_data(self):
        if self.private_key_data is None:
            return None

        private_key_password = None \
            if self.private_key_password is None \
            else self.private_key_password.encode()

        pkey = serialization.load_pem_private_key(self.private_key_data.encode(),
                                                  private_key_password,
                                                  backend=default_backend())

        return pkey.private_bytes(encoding=serialization.Encoding.DER,
                                  format=serialization.PrivateFormat.PKCS8,
                                  encryption_algorithm=serialization.NoEncryption())

    def get_column_definitions(self, schema, table, connection=None):
        columns = Table('COLUMNS', schema='INFORMATION_SCHEMA')

        columns_query = (
            SnowflakeQuery.from_(columns, immutable=False)
            .select(columns.COLUMN_NAME, columns.DATA_TYPE)
            .where(columns.TABLE_SCHEMA == Parameter('%(schema)s'))
            .where(columns.field('TABLE_NAME') == Parameter('%(table)s'))
            .distinct()
            .orderby(columns.column_name)
        )

        return self.fetch(str(columns_query), connection=connection, parameters=dict(schema=schema, table=table))



from functools import wraps
from multiprocessing.pool import ThreadPool


class ThreadPoolConcurrencyMiddleware:

    def __init__(self, max_processes=1):
        self.max_processes = max_processes

    def __call__(self, func):
        @wraps(func)
        def wrapper(database, *queries, **kwargs):
            with ThreadPool(processes=self.max_processes) as pool:
                results = pool.map(lambda query: func(database, query, **kwargs)[0], queries)
                pool.close()

            return results

        return wrapper



from .base import Database
from .column import (
    Column,
    ColumnsTransformer,
    make_columns,
)
from .mssql import MSSQLDatabase
from .mysql import (
    MySQLDatabase,
    MySQLTypeEngine,
)
from .postgresql import PostgreSQLDatabase
from .redshift import RedshiftDatabase
from .snowflake import SnowflakeDatabase
from .type_engine import TypeEngine
from .vertica import (
    VerticaDatabase,
    VerticaTypeEngine,
)



from pypika.terms import Term
from pypika.utils import format_alias_sql

from fireant.utils import (
    deepcopy,
    immutable,
)


class Modifier:
    wrapped_key = "wrapped"

    def __init__(self, wrapped):
        wrapped_key = super().__getattribute__("wrapped_key")
        setattr(self, wrapped_key, wrapped)

    def __getattr__(self, attr):
        if attr in ("__copy__", "__deepcopy__"):
            raise AttributeError()

        if attr in self.__dict__:
            return super().__getattribute__(attr)

        wrapped_key = super().__getattribute__("wrapped_key")
        wrapped = super().__getattribute__(wrapped_key)
        return getattr(wrapped, attr)

    def __setattr__(self, attr, value):
        wrapped_key = super().__getattribute__("wrapped_key")
        if attr == wrapped_key:
            super().__setattr__(attr, value)
            return

        wrapped = super().__getattribute__(wrapped_key)

        if attr in wrapped.__dict__:
            setattr(wrapped, attr, value)
            return

        super().__setattr__(attr, value)

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        wrapped_key = super().__getattribute__("wrapped_key")
        wrapped = super().__getattribute__(wrapped_key)
        return "{}({})".format(self.__class__.__name__, repr(wrapped))

    def __deepcopy__(self, memodict={}):
        wrapped_key = super().__getattribute__("wrapped_key")
        wrapped = super().__getattribute__(wrapped_key)
        memodict[id(wrapped)] = wrapped
        return deepcopy(self, memodict)

    @immutable
    def for_(self, wrapped):
        wrapped_key = super().__getattribute__("wrapped_key")
        setattr(self, wrapped_key, wrapped)


class FieldModifier:
    def __init__(self, field):
        self.field = field

    @immutable
    def for_(self, field):
        self.field = field

    def __deepcopy__(self, memodict={}):
        memodict[id(self.field)] = self.field
        return deepcopy(self, memodict)


class DimensionModifier(Modifier):
    """
    Base class for all dimension modifiers.
    """
    wrapped_key = "dimension"


class FilterModifier(Modifier):
    """
    Base class for all filter modifiers.
    """
    wrapped_key = "filter"

    @immutable
    def for_(self, field):
        self.filter.field = field


class RollupValue(Term):
    CONSTANT = "_FIREANT_ROLLUP_VALUE_"

    def get_sql(self, **kwargs):
        sql = f"'{self.CONSTANT}'"
        return format_alias_sql(sql, self.alias, **kwargs)


class Rollup(DimensionModifier):
    """
    A field modifier that will make totals be calculated for the wrapped dimension.
    """

    @property
    def definition(self):
        return RollupValue()


class OmitFromRollup(FilterModifier):
    """
    A filter modifier that will make the wrapped filter not apply for any total calculations, which might be
    available if an affected field has a `Rollup` dimension modifier set.
    """
    pass


class ResultSet(FilterModifier):
    """
    A filter modifier that will make the wrapped filter not filter the data. Instead, it will create new dimensions
    for representing membership of a set, given the wrapped filter's conditional.
    """
    def __init__(
        self, *args, set_label=None, complement_label=None, will_replace_referenced_dimension=True,
        will_group_complement=True, **kwargs
    ):
        """
        :param set_label: A string that will be used for naming the sub-set of the data, for which the
                          provided conditional evaluates to True (i.e the set itself). If not set,
                          a placeholder containing the conditional will be used instead.
        :param complement_label: A string that will be used for naming the sub-set of the data, for which the
                                 provided conditional evaluates to False (i.e the complement of the set). If not
                                 set, a placeholder containing the conditional will be used instead.
        :param will_replace_referenced_dimension: Whether a dimension referenced in the wrapped filter, if any,
                                                  should be replaced with the new dimension stating set membership.
                                                  If False, the wrapped filter's dimension will be kept as the
                                                  dimension right after the new set dimension. Defaults to True.
        :param will_group_complement: Whether the complement set should be broken down into its elements.
                                      Defaults to True.
        """
        super().__init__(*args, **kwargs)
        self.set_label = set_label
        self.complement_label = complement_label
        self.will_replace_referenced_dimension = will_replace_referenced_dimension
        self.will_group_complement = will_group_complement



import time
from functools import wraps

from fireant.middleware.slow_query_logger import (
    query_logger,
    slow_query_logger,
)


def log_middleware(func):

    @wraps(func)
    def wrapper(database, *queries, **kwargs):
        results = []

        for query in queries:
            start_time = time.time()
            query_logger.debug(query)

            results.append(func(database, query, **kwargs)[0])

            duration = round(time.time() - start_time, 4)
            query_log_msg = '[{duration} seconds]: {query}'.format(duration=duration,
                                                                   query=query)
            query_logger.info(query_log_msg)

            if database.slow_query_log_min_seconds is not None and duration >= database.slow_query_log_min_seconds:
                slow_query_logger.warning(query_log_msg)

        return results

    return wrapper


def connection_middleware(func):

    @wraps(func)
    def wrapper(database, *queries, **kwargs):
        connection = kwargs.pop('connection', None)
        if connection:
            return func(database, *queries, connection=connection, **kwargs)
        with database.connect() as connection:
            return func(database, *queries, connection=connection, **kwargs)

    return wrapper


def apply_middlewares(wrapped_func):

    @wraps(wrapped_func)
    def wrapper(database, *args, **kwargs):
        func = wrapped_func
        for middleware in reversed(database.middlewares):
            func = middleware(func)

        return func(database, *args, **kwargs)

    return wrapper



import copy
import csv
import inspect
import tempfile
from collections import OrderedDict
from functools import partial, wraps
from types import GeneratorType


def immutable(func):
    """
    Decorator for wrapper "builder" functions.  These are functions on the Query class or other classes used for
    building queries which mutate the query and return self.  To make the build functions immutable, this decorator is
    used which will deepcopy the current instance.  This decorator will return the return value of the inner function
    or the new copy of the instance.  The inner function does not need to return self.
    """

    def _copy(self, *args, mutate=False, **kwargs):
        """
        :param mutate:
            When True, overrides the immutable behavior of this decorator.
        """
        self_copy = self if mutate else copy.deepcopy(self)
        result = func(self_copy, *args, **kwargs)

        # Return self if the inner function returns None.  This way the inner function can return something
        # different (for example when creating joins, a different builder is returned).
        if result is None:
            return self_copy

        return result

    return _copy


def deepcopy(value, memodict):
    cls = value.__class__
    result = cls.__new__(cls)

    memodict[id(value)] = result

    for k, v in value.__dict__.items():
        result.__dict__[k] = copy.deepcopy(v, memodict)

    return result


def wrap_list(value, wrapper=list):
    return value if isinstance(value, (tuple, list)) else wrapper([value])


def setdeepattr(d, keys, value):
    """
    Similar to the built-in `setattr`, this function accepts a list/tuple of keys to set a value deep in a `dict`

    Given the following dict structure

    .. code-block:: python

        d = {
          'A': {
            '0': {
              'a': 1,
              'b': 2,
            }
          },
        }

    Calling `setdeepattr` with a key path to a value deep in the structure will set that value. If the value or any
    of the objects in the key path do not exist, then a dict will be created.

    .. code-block:: python

        # Overwrites the value in `d` at A.0.a, which was 1, to 3
        setdeepattr(d, ('A', '0', 'a'), 3)

        # Adds an entry in `d` to A.0 with the key 'c' and the value 3
        setdeepattr(d, ('A', '0', 'c'), 3)

        # Adds an entry in `d` with the key 'X' and the value a new dict
        # Adds an entry in `d` to `X` with the key '0' and the value a new dict
        # Adds an entry in `d` to `X.0` with the key 'a' and the value 0
        setdeepattr(d, ('X', '0', 'a'), 0)

    :param d:
        A dict value with nested dict attributes.
    :param keys:
        A list/tuple path of keys in `d` to the desired value
    :param value:
        The value to set at the given path `keys`.
    """
    if not isinstance(keys, (list, tuple)):
        keys = (keys,)

    top, *rest = keys

    if rest:
        if top not in d:
            d[top] = {}

        setdeepattr(d[top], rest, value)

    else:
        d[top] = value


def getdeepattr(d, keys, default_value=None):
    """
    Similar to the built-in `getattr`, this function accepts a list/tuple of keys to get a value deep in a `dict`

    Given the following dict structure

    .. code-block:: python

        d = {
          'A': {
            '0': {
              'a': 1,
              'b': 2,
            }
          },
        }

    Calling `getdeepattr` with a key path to a value deep in the structure will return that value. If the value or any
    of the objects in the key path do not exist, then the default value is returned.

    .. code-block:: python

        assert 1 == getdeepattr(d, ('A', '0', 'a'))
        assert 2 == getdeepattr(d, ('A', '0', 'b'))
        assert 0 == getdeepattr(d, ('A', '0', 'c'), default_value=0)
        assert 0 == getdeepattr(d, ('X', '0', 'a'), default_value=0)

    :param d:
        A dict value with nested dict attributes.
    :param keys:
        A list/tuple path of keys in `d` to the desired value
    :param default_value:
        A default value that will be returned if the path `keys` does not yield a value.
    :return:
        The value following the path `keys` or `default_value`
    """
    d_level = d

    for key in keys:
        if key not in d_level:
            return default_value

        d_level = d_level[key]

    return d_level


def apply_kwargs(f, *args, **kwargs):
    argspec = inspect.getfullargspec(f)
    allowed = set(argspec.args[-len(argspec.defaults or ()) :])
    return f(
        *args,
        **{
            key: kwarg
            for key, kwarg in kwargs.items()
            if argspec.varkw or key in allowed
        }
    )


def filter_kwargs(f):
    """
    Removes any kwargs from function call that are not accepted by the called function.

    :param f:
    :return:
    """
    return partial(apply_kwargs, f)


def flatten(items):
    return [item for level in items for item in wrap_list(level)]


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    return [l[i : i + n] for i in range(0, len(l), n)]


def ordered_distinct_list(l):
    seen = set()
    return [x for x in l if not x in seen and not seen.add(x)]


def ordered_distinct_list_by_attr(l, attr="alias", keep_last=False):
    mapping = {}  # will contain mapping from attr_value to item
    for item in l:
        attr_value = getattr(item, attr)
        if not keep_last and attr_value in mapping:
            # only keep the first occurrence in the result so skip the item if the attribute value is already present
            continue

        mapping[attr_value] = item

    return list(mapping.values())


def groupby(items, by):
    """
    Group items using a function to derive a key.

    :param items:
        The items to group

    :param by:
        A lambda function to create a key based on the item

    :return:
        an Ordered dict
    """
    result = OrderedDict()
    for item in items:
        key = by(item)

        if key in result:
            result[key].append(item)
        else:
            result[key] = [item]

    return result


def alias_selector(alias):
    if alias is None or alias.startswith("$"):
        return alias
    return "${}".format(alias)


def alias_for_alias_selector(f_alias):
    if f_alias and f_alias[0] == "$":
        return f_alias[1:]
    return f_alias


def reduce_data_frame_levels(data_frame, level):
    reduced = data_frame.reset_index(level=level, drop=True)
    if reduced.size == 1 and reduced.index.names == [None]:
        return reduced.iloc[0]
    return reduced


def read_csv(fp):
    """
    Read a csv file and return its content.

    :param fp: Path of the file.
    :return: List of rows.
    """
    rows = []
    with open(fp, "r") as f:
        csv_reader = csv.reader(f, delimiter=",")
        for row in csv_reader:
            rows.append(row)

    return rows


def write_named_temp_csv(rows):
    """
    Write the provided rows to a named temporary file.

    :param rows: List of lists to be written.
    :return: A named temporary file containing the rows.
    """
    ntf = tempfile.NamedTemporaryFile(suffix=".csv")
    with open(ntf.name, "w") as f:
        ntf_writer = csv.writer(f, delimiter=",")
        ntf_writer.writerows(rows)

    return ntf


def listify(func):
    @wraps(func)
    def new_func(*args, **kwargs):
        retval = func(*args, **kwargs)
        if isinstance(retval, GeneratorType):
            return list(retval)
        return retval

    return new_func



from pypika import (
    Parameter, Tables,
    VerticaQuery,
    functions as fn,
    terms,
)

from .base import Database
from .sql_types import (
    BigInt,
    Boolean,
    Char,
    Date,
    DateTime,
    Decimal,
    DoublePrecision,
    Float,
    Integer,
    Numeric,
    Real,
    SmallInt,
    Text,
    Time,
    Timestamp,
    VarChar,
)
from .type_engine import TypeEngine


class Trunc(terms.Function):
    """
    Wrapper for Vertica TRUNC function for truncating dates.
    """

    def __init__(self, field, date_format, alias=None):
        super(Trunc, self).__init__('TRUNC', field, date_format, alias=alias)


class VerticaDatabase(Database):
    """
    Vertica client that uses the vertica_python driver.
    """

    # The pypika query class to use for constructing queries
    query_cls = VerticaQuery

    DATETIME_INTERVALS = {
        "hour": "HH",
        "day": "DD",
        "week": "IW",
        "month": "MM",
        "quarter": "Q",
        "year": "Y",
    }

    def __init__(
        self,
        host="localhost",
        port=5433,
        database="vertica",
        user="vertica",
        password=None,
        read_timeout=None,
        **kwags
    ):
        super(VerticaDatabase, self).__init__(host, port, database, **kwags)
        self.user = user
        self.password = password
        self.read_timeout = read_timeout
        self.type_engine = VerticaTypeEngine()

    def connect(self):
        import vertica_python

        return vertica_python.connect(
            host=self.host,
            port=self.port,
            database=self.database,
            user=self.user,
            password=self.password,
            read_timeout=self.read_timeout,
            unicode_error="replace",
        )

    def trunc_date(self, field, interval):
        trunc_date_interval = self.DATETIME_INTERVALS.get(str(interval), "DD")
        return Trunc(field, trunc_date_interval)

    def date_add(self, field, date_part, interval):
        return fn.TimestampAdd(str(date_part), interval, field)

    def get_column_definitions(self, schema, table, connection=None):
        view_columns, table_columns = Tables('view_columns', 'columns')

        view_query = VerticaQuery.from_(view_columns) \
            .select(view_columns.column_name, view_columns.data_type) \
            .where((view_columns.table_schema == Parameter(':schema'))
                   & (view_columns.field('table_name') == Parameter(':table'))) \
            .distinct()

        table_query = (
            VerticaQuery.from_(table_columns, immutable=False)
            .select(table_columns.column_name, table_columns.data_type)
            .where(
                (table_columns.table_schema == Parameter(':schema'))
                & (table_columns.field("table_name") == Parameter(':table'))
            )
            .distinct()
        )

        return self.fetch(
            str(view_query + table_query),
            parameters=dict(schema=schema, table=table),
            connection=connection
        )

    def import_csv(self, table, file_path, connection=None):
        """
        Imports a file into a database table.

        :param table: The name of a table to import data into.
        :param file_path: The path of the file to be imported.
        :param connection: (Optional) The connection to execute this query with.
        """
        import_query = VerticaQuery.from_file(file_path).copy_(table)

        self.execute(str(import_query), connection=connection)

    def create_temporary_table_from_columns(self, table, columns, connection=None):
        """
        Creates a temporary table from a list of columns.

        :param table: The name of the new temporary table.
        :param columns: The columns of the new temporary table.
        :param connection: (Optional) The connection to execute this query with.
        """
        create_query = (
            VerticaQuery.create_table(table)
            .temporary()
            .local()
            .preserve_rows()
            .columns(*columns)
        )

        self.execute(str(create_query), connection=connection)

    def create_temporary_table_from_select(self, table, select_query, connection=None):
        """
        Creates a temporary table from a SELECT query.

        :param table: The name of the new temporary table.
        :param select_query: The query to be used for selecting data of an existing table for the new temporary table.
        :param connection: (Optional) The connection to execute this query with.
        """
        create_query = (
            VerticaQuery.create_table(table)
            .temporary()
            .local()
            .preserve_rows()
            .as_select(select_query)
        )

        self.execute(str(create_query), connection=connection)


class VerticaTypeEngine(TypeEngine):
    vertica_to_ansi_mapper = {
        "char": Char,
        "varchar": VarChar,
        "varchar2": VarChar,
        "longvarchar": Text,
        "boolean": Boolean,
        "int": Integer,
        "integer": Integer,
        "int8": Integer,
        "smallint": SmallInt,
        "tinyint": SmallInt,
        "bigint": BigInt,
        "decimal": Decimal,
        "numeric": Numeric,
        "number": Numeric,
        "float": Float,
        "float8": Float,
        "real": Real,
        "double": DoublePrecision,
        "date": Date,
        "time": Time,
        "timetz": Time,
        "datetime": DateTime,
        "smalldatetime": DateTime,
        "timestamp": Timestamp,
        "timestamptz": Timestamp,
    }

    ansi_to_vertica_mapper = {
        "CHAR": "char",
        "VARCHAR": "varchar",
        "TEXT": "longvarchar",
        "BOOLEAN": "boolean",
        "INTEGER": "integer",
        "SMALLINT": "smallint",
        "BIGINT": "bigint",
        "DECIMAL": "decimal",
        "NUMERIC": "numeric",
        "FLOAT": "float",
        "REAL": "real",
        "DOUBLEPRECISION": "double",
        "DATE": "date",
        "TIME": "time",
        "DATETIME": "datetime",
        "TIMESTAMP": "timestamp",
    }

    def __init__(self):
        super(VerticaTypeEngine, self).__init__(
            self.vertica_to_ansi_mapper, self.ansi_to_vertica_mapper
        )



import sys

import numpy as np
import pandas as pd

from .modifiers import Rollup

DATE_TOTALS = pd.Timestamp.max
NUMBER_TOTALS = sys.maxsize
TEXT_TOTALS = '~~totals'

TOTALS_MARKERS = {TEXT_TOTALS, NUMBER_TOTALS, DATE_TOTALS}


def get_totals_marker_for_dtype(dtype):
    """
    For a given dtype, return the index value to use to indicate that the data frame row contains totals.

    :param dtype:
    :return:
    """
    return {
        np.dtype('<M8[ns]'): DATE_TOTALS,
        np.dtype('int64'): NUMBER_TOTALS,
        np.dtype('float64'): NUMBER_TOTALS,
    }.get(dtype, TEXT_TOTALS)


def scrub_totals_from_share_results(data_frame, dimensions):
    """
    This function returns a data frame with the values for dimension totals filtered out if the corresponding dimension
    was not queried with rollup. This comes into play when the share operation is used on metrics which requires the
    totals across the values.

    There are two different logical branches for this function that perform the same work, one on a single-level index
    data frame and another for a multi-level index.

    :param data_frame:
        The result data set.
    :param dimensions:
        A list of dimensions that were queried for to produce the result data set.
    :return:
        The data frame with totals rows removed for dimensions that were not queried with rollup.
    """
    if isinstance(data_frame.index, pd.MultiIndex):
        return _scrub_totals_for_multilevel_index_df(data_frame, dimensions)
    return _scrub_totals_for_singlelevel_index_df(data_frame, dimensions)


def _scrub_totals_for_singlelevel_index_df(data_frame, dimensions):
    # For this function, there will only ever be zero or 1 dimensions in the list of dimensions.
    # If there are zero dimensions or the only dimension is rolled up, then return the data frame as-is
    is_rolled_up = dimensions and isinstance(dimensions[0], Rollup)
    if is_rolled_up:
        return data_frame

    # Otherwise, remove any rows where the index value equals the totals marker for its dtype.
    marker = get_totals_marker_for_dtype(data_frame.index.dtype)
    is_totals = data_frame.index == marker
    return data_frame[~is_totals]


def _scrub_totals_for_multilevel_index_df(data_frame, dimensions):
    if data_frame.empty:
        return data_frame

    # Get the totals marker value for each index level
    markers = [get_totals_marker_for_dtype(level.dtype)
               for level in data_frame.index.levels]

    # Create a boolean data frame indicating whether or not the index value equals the totals marker for the dtype
    # corresponding to each index level
    is_total_marker = pd.DataFrame([[value == marker
                                     for value, marker in zip(values, markers)]
                                    for values in data_frame.index],
                                   index=data_frame.index)

    """
    If a row in the data frame is for totals for one index level, all of the subsequent index levels will also use a
    totals marker. In order to avoid filtering the wrong rows, a new data frame is created similar to `is_totals_marker`
    except a cell is only set to True if that value is a totals marker for the corresponding index level, the leaves of
    the dimension value tree.

    This is achieved by rolling an XOR function across each index level with the previous level.
    """
    first_column = is_total_marker.columns[0]
    is_totals_marker_leaf = pd.DataFrame(is_total_marker[first_column])
    for column, prev_column in zip(is_total_marker.columns[1:], list(is_total_marker.columns[:-1])):
        is_totals_marker_leaf[column] = np.logical_xor(is_total_marker[column], is_total_marker[prev_column])

    # Create a boolean vector for each dimension to mark if that dimension is rolled up
    rollup_dimensions = np.array([isinstance(dimension, Rollup)
                                  for dimension in dimensions])

    # Create a boolean pd.Series where False means to remove the row from the data frame.
    mask = (~(~rollup_dimensions & is_totals_marker_leaf)).all(axis=1)
    return data_frame.loc[mask]



class Annotation(object):
    def __init__(self, table, field, alignment_field, dataset_alignment_field_alias):
        """
        :param table:
            A Pypika instance of the annotation table.
        :param field:
            The annotation field.
        :param alignment_field:
            An additional field in the annotation table for aligning the annotation data with the dataset.
        :param dataset_alignment_field_alias:
            An alias of the alignment dimension in the associated dataset.
        """
        self.table = table
        self.field = field
        self.alignment_field = alignment_field
        self.dataset_alignment_field_alias = dataset_alignment_field_alias



from datetime import datetime
from typing import Collection, Dict, Union

import pandas as pd
from pypika import (
    Query,
    enums,
    functions as fn,
    terms,
)
from pypika.terms import Function

from fireant.middleware.decorators import (apply_middlewares, connection_middleware)


class Database(object):
    """
    This is a abstract base class used for interfacing with a database platform.
    """

    # The pypika query class to use for constructing queries
    query_cls = Query

    slow_query_log_min_seconds = 15
    can_group_static_value = True

    def __init__(
        self,
        host=None,
        port=None,
        database=None,
        max_result_set_size=200000,
        middlewares=[],
    ):
        self.host = host
        self.port = port
        self.database = database
        self.max_result_set_size = max_result_set_size
        self.middlewares = middlewares + [connection_middleware]

    def connect(self):
        """
        This function must establish a connection to the database platform and return it.
        """
        raise NotImplementedError

    def get_column_definitions(self, schema, table, connection=None):
        """
        Return a list of column name, column data type pairs.

        :param schema: The name of the table schema.
        :param table: The name of the table to get columns from.
        :param connection: (Optional) The connection to execute this query with.
        :return: A list of columns.
        """
        raise NotImplementedError

    def trunc_date(self, field, interval):
        """
        This function must create a Pypika function which truncates a Date or DateTime object to a specific interval.
        """
        raise NotImplementedError

    def date_add(self, field: terms.Term, date_part: str, interval: int):
        """
        This function must add/subtract a Date or Date/Time object.
        """
        raise NotImplementedError

    def convert_date(self, dt: datetime) -> Union[datetime, Function]:
        """
        Override to provide a custom function for converting a date.
        Defaults to an identity function.

        :param dt: Date to convert
        """
        return dt

    def to_char(self, definition):
        return fn.Cast(definition, enums.SqlTypes.VARCHAR)

    @apply_middlewares
    def fetch_queries(self, *queries, connection=None, parameters: Union[Dict, Collection] = ()):
        results = []
        # Parameters can either be passed as a list when using formatting placeholders like %s (varies per platform)
        # or a dict when using named placeholders.
        for query in queries:
            cursor = connection.cursor()
            cursor.execute(str(query), parameters)
            results.append(cursor.fetchall())

        return results

    def fetch(self, query, **kwargs):
        return self.fetch_queries(query, **kwargs)[0]

    @apply_middlewares
    def execute(self, *queries, **kwargs):
        connection = kwargs.get("connection")
        for query in queries:
            cursor = connection.cursor()
            cursor.execute(str(query))
            connection.commit()

    @apply_middlewares
    def fetch_dataframes(self, *queries, parse_dates=None, **kwargs):
        connection = kwargs.get("connection")
        dataframes = []
        for query in queries:
            dataframes.append(
                pd.read_sql(query, connection, coerce_float=True, parse_dates=parse_dates)
            )
        return dataframes

    def fetch_dataframe(self, query, **kwargs):
        return self.fetch_dataframes(query, **kwargs)[0]

    def __str__(self):
        return f'Database|{self.__class__.__name__}|{self.host}'



from typing import Tuple

import pandas as pd
from pandas.core.dtypes.common import is_datetime64_ns_dtype
from pypika import Order

from fireant.utils import alias_selector


def _get_window(limit, offset):
    start = offset
    end = offset + limit if None not in (offset, limit) else limit
    return start, end


def _get_sorting_schema(orders) -> Tuple[list, bool]:
    sort_values, ascending = zip(
        *[
            (alias_selector(field.alias), orientation is Order.asc)
            for field, orientation in orders
        ]
    )
    return list(sort_values), ascending


def paginate(data_frame, widgets, orders=(), limit=None, offset=None):
    """
    :param data_frame:
        The result set to sort.
    :param widgets:
        An iterable of widgets that the sort is being applied for.
    :param orders:
        An iterable of (<Dimension/Metric>, pypika.Order)
     :param limit:
        A limit of the number of data points/series
    :param offset:
        A offset of the number of data points/series
    :return:
        A paginated data frame. If the widget required grouped pagination, then there should be an upperbound
        `limit*(n_index_level_0)`. Otherwise the data frame should have the same length as the limit.
    :return:
        A sorted data frame.
    """
    if len(data_frame) == 0:
        return data_frame

    start, end = _get_window(limit, offset)

    needs_group_pagination = isinstance(data_frame.index, pd.MultiIndex) and any(
        [getattr(widget, 'group_pagination', False) for widget in widgets]
    )

    if needs_group_pagination:
        return _group_paginate(data_frame, start, end, orders)
    return _simple_paginate(data_frame, start, end, orders)


def _simple_paginate(data_frame, start=None, end=None, orders=()):
    """
    Applies pagination which limits the number of rows in the dataframe.

    :param data_frame:
        A data frame to paginate
    :param start:
        The index starting point to slice the data frame at
    :param end:
        The index ending point to slice the data frame at
    :param orders:
        A list of tuples that contain a slicer field definition (with an alias matching the columns of the data frame)
        and a pypika.Order.
    """
    if orders:
        sort, ascending = _get_sorting_schema(orders)
        data_frame = data_frame.sort_values(by=sort, ascending=ascending)

    return data_frame[start:end]


def _index_isnull(data_frame):
    if isinstance(data_frame.index, pd.MultiIndex):
        return [
            any(pd.isnull(value) for value in level) for level in list(data_frame.index)
        ]

    return pd.isnull(data_frame.index)


def _aggregate_dimension_groups(group):
    # FIXME this should aggregate according to field definition, instead of sum/max
    # Need a way to interpret definitions in python code in order to do that
    if is_datetime64_ns_dtype(group):
        # sum aggregation doesn't work on the datetime type so use max instead
        return group.max()
    return group.sum()


def _group_paginate(data_frame, start=None, end=None, orders=()):
    """
    Applies pagination which limits the number of rows in the data frame grouped by the zeroth index level. This will
    in turn paginate the number of series in the data frame.
    :param data_frame:
        A data frame to paginate
    :param start:
        The index starting point to slice the data frame at
    :param end:
        The index ending point to slice the data frame at
    :param orders:
        A list of tuples that contain a slicer field definition (with an alias matching the columns of the data frame)
        and a pypika.Order.
    """
    dimension_levels = data_frame.index.names[1:]
    dimension_groups = data_frame.groupby(level=dimension_levels)

    # Do not apply ordering on the 0th dimension !!!
    # This would not have any result since the X-Axis on a chart is ordered sequentially
    orders = [
        (field, orientation)
        for field, orientation in orders
        if alias_selector(field.alias) != data_frame.index.names[0]
    ]

    if orders:
        aggregated_df = dimension_groups.aggregate(_aggregate_dimension_groups)
        sort, ascending = _get_sorting_schema(orders)
        sorted_df = aggregated_df.sort_values(by=sort, ascending=ascending)
        sorted_dimension_values = tuple(sorted_df.index)[start:end]

    else:
        sorted_dimension_values = tuple(dimension_groups.apply(lambda g: g.name))[start:end]

    sorted_dimension_values = (
        pd.Index(sorted_dimension_values, name=dimension_levels[0])
        if len(dimension_levels) == 1
        else pd.MultiIndex.from_tuples(sorted_dimension_values, names=dimension_levels)
    )

    def _apply_pagination(df):
        """
        This function applies sorting by using the sorted dimension values as an index to select values in the right
        order out of the data frame. The index must be filtered to only values that are in this data frame, since
        there might be missing combinations of index values.
        """
        dfx = df.reset_index(level=0, drop=True)
        value_in_index = sorted_dimension_values.isin(dfx.index)
        index_slice = sorted_dimension_values[value_in_index].values

        """
        In the case of bool dimensions, convert index_slice to an array of literal `True` and/or `False`,
        because pandas `.loc` handles lists of bool as a mask. In which `True` returns the row and `False` won't.
        """
        if bool in {type(x) for x in sorted_dimension_values}:
            num_missing_entries_in_boolean_mask = len(index_slice)
            index_slice = [False]*len(dfx.index)

            for i in range(num_missing_entries_in_boolean_mask):
                index_slice[i] = True

        # Need to include nulls so append them to the end of the sorted data frame
        isnull = _index_isnull(dfx)

        return dfx.loc[index_slice, :].append(dfx[isnull])

    return (
        data_frame.sort_values(data_frame.index.names[0], ascending=True)
        .groupby(level=0)
        .apply(_apply_pagination)
    )



from typing import TYPE_CHECKING, Union

from pypika import Order

from fireant.dataset.fields import Field
from fireant.exceptions import DataSetException
from fireant.utils import (
    deepcopy,
    immutable,
)
from ..execution import fetch_data
from ..finders import find_field_in_modified_field
from ..sets import (
    apply_set_dimensions,
    omit_set_filters,
)

if TYPE_CHECKING:
    from fireant.dataset import DataSet


class QueryException(DataSetException):
    pass


def add_hints(queries, hint=None):
    return [
        query.hint(hint)
        if hint is not None and hasattr(query.__class__, "hint")
        else query
        for query in queries
    ]


def get_column_names(database, table):
    column_definitions = database.get_column_definitions(
        table._schema._name, table._table_name
    )

    return {column_definition[0] for column_definition in column_definitions}


def validate_fields(fields, dataset):
    fields = [find_field_in_modified_field(field) for field in fields]

    invalid = [field.alias for field in fields if field not in dataset.fields]
    if not invalid:
        return

    raise DataSetException(
          "Only fields from dataset can be used in a dataset query. Found invalid fields: {}.".format(
                ", ".join(invalid)
          )
    )


def _strip_modifiers(fields):
    for field in fields:
        node = field
        while hasattr(node, "dimension"):
            node = node.dimension
        yield node


class QueryBuilder(object):
    """
    This is the base class for building dataset queries. This class provides an interface for building dataset queries
    via a set of functions which can be chained together.
    """

    def __init__(self, dataset: 'DataSet'):
        """
        :param dataset: DataSet to build the query for
        """
        self.dataset = dataset
        self.table = dataset.table
        self._dimensions = []
        self._filters = []
        self._orders = None
        self._client_limit = None
        self._client_offset = None
        self._query_limit = None
        self._query_offset = None

    # noinspection PyDefaultArgument
    def __deepcopy__(self, memodict={}):
        fields = [d for d in self._dimensions]
        if self._orders is not None:
            fields += [field for (field, _) in self._orders]

        for field in fields:
            field = find_field_in_modified_field(field)
            memodict[id(field)] = field

        return deepcopy(self, memodict)

    @immutable
    def dimension(self, *dimensions):
        """
        Add one or more dimensions when building a dataset query.

        :param dimensions:
            Dimensions to add to the query
        :return:
            A copy of the query with the dimensions added.
        """
        validate_fields(dimensions, self.dataset)
        aliases = {dimension.alias for dimension in self._dimensions}
        self._dimensions += [
            dimension for dimension in dimensions if dimension.alias not in aliases
        ]

    @immutable
    def filter(self, *filters):
        """
        Add one or more filters when building a dataset query.

        :param filters:
            Filters to add to the query
        :return:
            A copy of the query with the filters added.
        """
        validate_fields([fltr.field for fltr in filters], self.dataset)
        self._filters += [f for f in filters]

    @immutable
    def orderby(self, field: Field, orientation: Order = None):
        """
        :param field:
            The element to order by, either a metric or dimension.
        :param orientation:
            The directionality to order by, either ascending or descending.
        :return:
            A copy of the query with the order by added.
        """
        validate_fields([field], self.dataset)

        if self._orders is None:
            self._orders = []

        if field is not None:
            self._orders += [(field, orientation)]

    @immutable
    def limit_query(self, limit):
        """
        Sets the limit of the query.

        :param limit:
            A limit on the number of database rows returned.
        :return:
            A copy of the query with the query limit set.
        """
        self._query_limit = limit

    @immutable
    def offset_query(self, offset):
        """
        Sets the offset of the query.

        :param offset:
            A offset on the number of database rows returned.
        :return:
            A copy of the query with the query offset set.
        """
        self._query_offset = offset

    @immutable
    def limit_client(self, limit):
        """
        Sets the limit to be used when paginating the Pandas Dataframe after the results
        have been returned

        :param limit:
            A limit on the number of dataframe rows returned.
        :return:
            A copy of the query with the dataframe limit set.
        """
        self._client_limit = limit

    @immutable
    def offset_client(self, offset):
        """
        Sets the offset to be used when paginating the Pandas Dataframe after the results
        have been returned

        :param offset:
            A offset on the number of dataframe rows returned
        :return:
            A copy of the query with the dataframe offset set.
        """
        self._client_offset = offset

    @property
    def dimensions(self):
        """
        Returns a list of Field instances, that might include newly created set dimensions.

        Set dimensions are generated at this level because the `ResultSet` filter modifier can artificially
        create dimensions, which widgets need to be aware in order to properly render the data. Moving this to
        `make_slicer_query_with_totals_and_references` function is not possible, even if we defaulted to using
        same alias as the referenced dimension/metric, given that would cause aliases clashes. Dimensions can
        be mostly replaced, but that's not the case for metrics.

        :return: A list of Field instances.
        """
        return apply_set_dimensions(self._dimensions, self._filters, self.dataset)

    @property
    def filters(self):
        """
        Returns a list of Filter instances, that might omit filters wrapped with `ResultSet` filter modifier.

        :return: A list of Filter instances.
        """
        return omit_set_filters(self._filters)

    @property
    def orders(self):
        """
        Return orders.

        :return: None or a list of tuples shaped as Field instance and ordering.
        """
        return self._orders or self.default_orders

    @property
    def default_orders(self):
        """
        Return orders based on the provided dimensions.

        :return: A list of tuples shaped as Field instance and ordering.
        """
        dimension_orders = []

        for dimension in self.dimensions:
            if not dimension.is_aggregate:
                dimension_orders.append((dimension, None))

        return dimension_orders

    @property
    def sql(self):
        """
        Serialize this query builder object to a set of Pypika/SQL queries.

        This is the base implementation shared by two implementations: the query to fetch data for a dataset request and
        the query to fetch choices for dimensions.

        This function only handles dimensions (select+group by) and filtering (where/having), which is everything needed
        for the query to fetch choices for dimensions.

        The dataset query extends this with metrics, references, and totals.
        """
        raise NotImplementedError()

    def fetch(self, hint=None):
        """
        Fetches the data for this query instance and returns it in an instance of `pd.DataFrame`

        :param hint:
            For database vendors that support it, add a query hint to collect analytics on the queries triggered by
            fireant.
        """
        queries = add_hints(self.sql, hint)

        max_rows_returned, data = fetch_data(self.dataset.database, queries, self.dimensions)
        return self._transform_for_return(data, max_rows_returned=max_rows_returned)

    def _apply_pagination(self, query):
        # Some platforms require an order by when pagination is used. Therefore, if there is no ordering set,
        # we just default to the first column.
        if not self.orders:
            query = query.orderby(1)

        query = query.limit(min(self._query_limit or float('inf'), self.dataset.database.max_result_set_size))
        return query.offset(self._query_offset)

    def _transform_for_return(self, widget_data, **metadata) -> Union[dict, list]:
        return dict(data=widget_data, metadata=dict(**metadata)) \
            if self.dataset.return_additional_metadata \
            else widget_data


class ReferenceQueryBuilderMixin:
    """
    This is a mixin class for building dataset queries that allow references. This class provides an interface for
    building dataset queries via a set of functions which can be chained together.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._references = []

    @immutable
    def reference(self, *references):
        """
        Add one or more references for a dimension when building a dataset query.

        :param references:
            References to add to the query
        :return:
            A copy of the query with the references added.
        """
        validate_fields([reference.field for reference in references], self.dataset)
        self._references += references


class WidgetQueryBuilderMixin:
    """
    This is a mixin class for building dataset queries that allow widgets. This class provides an interface for
    building dataset queries via a set of functions which can be chained together.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._widgets = []

    def _validate(self):
        for widget in self._widgets:
            if hasattr(widget, "validate"):
                widget.validate(self._dimensions)

    @immutable
    def widget(self, *widgets):
        """
        Add one or more widgets when building a dataset query.

        :param widgets:
            Widgets to add to the query
        :return:
            A copy of the query with the widgets added.
        """
        validate_fields(
              [field for widget in widgets for field in widget.metrics], self.dataset
        )

        self._widgets += widgets



import copy
from typing import List

from pypika import JoinType, Query

from fireant.queries.builder.dataset_query_builder import DataSetQueryBuilder
from fireant.queries.finders import (
    find_dataset_fields,
    find_field_in_modified_field,
    find_metrics_for_widgets,
    find_operations_for_widgets,
    find_share_dimensions,
)
from fireant.queries.sql_transformer import make_slicer_query_with_totals_and_references
from fireant.reference_helpers import reference_type_alias
from fireant.utils import alias_selector, listify, ordered_distinct_list_by_attr
from fireant.widgets.base import Widget
from ..sets import (
    apply_set_dimensions,
    omit_set_filters,
)


@listify
def _find_dataset_fields_needed_to_be_mapped(dataset):
    """
    This produces a list of fields from a DatasetBlender that need to be mapped up to the Datasets. This is any simple
    fields, or fields that are just pointers to fields in a Dataset, and breaks up complex fields, fields that are
    defined with expressions referencing simple fields from the datasets, down to those fields.

    Blender
        primary
            DS0
                Fields
                    dimension0
                    metric0
        secondary
            DS1
                Fields
                    dimension0
                    metric1
        Fields
            dimension0  (points to dimension0 in DS0 and DS1)
            metric0  (points to metric0 in DS0)
            metric1  (points to metric0 in DS1)
            complex_metric DS0.metric0/DS1.metric1 (Composed of fields metric0 and metric1)

    Calling this function for each of the following will yield
    DS0 -> [DS0.dimension0, DS0.metric0]
    DS1 -> [DS1.dimension0, DS1.metric1]
    Blender -> [Blender.dimension0, Blender.metric0, Blender.metric1, DS0.metric0, DS1.metric1]
        (the Blender has a reference to each field in each DS and also a complex metric composed of metric0 and metric1)
    """
    complex_fields = []
    for field in dataset.fields:
        if field.is_wrapped:
            yield field
        else:
            complex_fields.append(field)

    yield from find_dataset_fields(complex_fields)


@listify
def _map_field(dataset_fields, blender_dataset_fields, dimension_map=None):
    """
    Returns a list of tuples shaped as blended dataset field and its matching field in the provided fields. Blender
    fields that have to match in the dataset fields will be ignored. Please note that the Field class has a custom
    dunder hash method.

    :param dataset_fields: A set of Field instances pertaining to a single dataset (i.e. primary or secondary).
    :param blender_dataset_fields: A list of Field instances in the blender dataset.
    :param dimension_map: A dict mapping Field instances from different datasets.
                          Often used to express the equivalency of primary and secondary dataset fields.
    :return: A list of tuples shaped as blended dataset field and its matching field in the provided fields.
    """
    for field in blender_dataset_fields:
        if dimension_map is not None and field.definition in dimension_map:
            yield field, dimension_map[field.definition]
        if field.definition in dataset_fields:
            yield field, field.definition
            # also yield the dataset field mapped to itself so that any reference to this field (from blender or
            # from dataset) can be used.
            yield field.definition, field.definition
        if field in dataset_fields:
            yield field, field


def _datasets_and_field_maps(blender_dataset, filters):
    """
    Returns a list of tuples shaped as dataset and its fields mapped per the blender dataset ones. The field maps
    are later used for knowing which columns to query on the respective primary and secondary datasets, given the
    columns selected for the blender dataset.

    :param blender_dataset: A DataSetBlender instance.
    :param filters: A list of Filter instances, present on the blender query. This is used for creating dimensions
                    derived from filters, such as the set dimension.
    :return: A list of tuples shaped as dataset and its fields mapped per the blender dataset ones.
    """
    from fireant.dataset.data_blending import DataSetBlender

    def _flatten_blend_datasets(dataset) -> List:
        primary_dataset = dataset.primary_dataset
        secondary_dataset = dataset.secondary_dataset
        blender_dataset_fields = apply_set_dimensions(
            _find_dataset_fields_needed_to_be_mapped(dataset), filters, dataset,
        )

        primary_dataset_fields = set(apply_set_dimensions(primary_dataset.fields, filters, primary_dataset))
        secondary_dataset_fields = set(apply_set_dimensions(secondary_dataset.fields, filters, secondary_dataset))

        blender_field_to_primary_field_map = dict(_map_field(
            primary_dataset_fields, blender_dataset_fields,
        ))

        primary_mapped_dimensions = list(
            apply_set_dimensions(dataset.dimension_map.keys(), filters, primary_dataset)
        )
        secondary_mapped_dimensions = list(
            apply_set_dimensions(dataset.dimension_map.values(), filters, secondary_dataset)
        )
        secondary_dimension_per_primary_dimension = dict(zip(
            primary_mapped_dimensions, secondary_mapped_dimensions,
        ))

        blender_field_to_secondary_field_map = dict(
            _map_field(secondary_dataset_fields, blender_dataset_fields, secondary_dimension_per_primary_dimension)
        )

        if not isinstance(primary_dataset, DataSetBlender):
            return [
                (primary_dataset, blender_field_to_primary_field_map),
                (secondary_dataset, blender_field_to_secondary_field_map),
            ]

        # get the dataset children of the blender (`dataset.primary_dataset`) and their corresponding field_maps,
        # then update the field map to reference this blender's field (`dataset`)
        datasets_and_field_maps = []
        for ds, fm in _flatten_blend_datasets(primary_dataset):
            remapped_field_map = {**fm}
            for field in blender_dataset_fields:
                if (
                    field not in blender_field_to_primary_field_map
                    or blender_field_to_primary_field_map[field] not in fm
                ):
                    continue
                remapped_field_map[field] = fm[blender_field_to_primary_field_map[field]]

            datasets_and_field_maps.append((ds, remapped_field_map))

        return [*datasets_and_field_maps, (secondary_dataset, blender_field_to_secondary_field_map)]

    return zip(*_flatten_blend_datasets(blender_dataset))


class EmptyWidget(Widget):
    @property
    def metrics(self):
        if 0 == len(self.items):
            return []

        return super().metrics


def _build_dataset_query(
    dataset, field_map, metrics, dimensions, filters, references, operations, share_dimensions
):
    @listify
    def _map_fields(fields):
        """
        TODO describe this
        """
        for field in fields:
            field_from_blender = find_field_in_modified_field(field)
            if field_from_blender in dataset.fields:
                yield field
                continue
            if field_from_blender not in field_map:
                continue

            yield field.for_(field_map[field_from_blender])

    dataset_metrics = ordered_distinct_list_by_attr(_map_fields(metrics))
    dataset_dimensions = _map_fields(dimensions)
    dataset_filters = _map_fields(omit_set_filters(filters))
    dataset_references = _map_fields(references)
    dataset_share_dimensions = _map_fields(share_dimensions)

    if not any([dataset_metrics, dataset_dimensions]):
        return [None]

    # TODO: It's possible that we have to adapt/map the operations for @apply_special_cases
    dataset_operations = operations

    return make_slicer_query_with_totals_and_references(
        database=dataset.database,
        table=dataset.table,
        joins=dataset.joins,
        dimensions=dataset_dimensions,
        metrics=dataset_metrics,
        operations=dataset_operations,
        filters=dataset_filters,
        references=dataset_references,
        orders=[],
        share_dimensions=dataset_share_dimensions,
    )


def _blender_join_criteria(
    base_query, join_query, dimensions, base_field_map, join_field_map
):
    """
    Build a criteria for joining this join query to the base query in dataset blender queries. This should be a set of
    equality conditions like A0=B0 AND A1=B1 AND An=Bn for each mapped dimension between dataset from
    `DataSetBlender.dimension_map`.
    """
    join_criteria = None
    for dimension in dimensions:
        dimension = find_field_in_modified_field(dimension)
        # dimension has to be in both field maps:
        if dimension not in base_field_map or dimension not in join_field_map:
            continue

        alias0, alias1 = [
            alias_selector(field_map[dimension].alias)
            for field_map in [base_field_map, join_field_map]
        ]

        next_criteria = base_query[alias0] == join_query[alias1]
        join_criteria = (
            next_criteria if join_criteria is None else (join_criteria & next_criteria)
        )

    return join_criteria


def _get_sq_field_for_blender_field(field, queries, field_maps, reference=None):
    unmodified_field = find_field_in_modified_field(field)
    field_alias = alias_selector(reference_type_alias(field, reference))

    # search for the field in each field map to determine which subquery it will be in
    for query, field_map in zip(queries, field_maps):
        if query is None or unmodified_field not in field_map:
            continue

        mapped_field = field_map[unmodified_field]
        mapped_field_alias = alias_selector(
            reference_type_alias(mapped_field, reference)
        )

        subquery_field = query[mapped_field_alias]
        # case #1 modified fields, ex. day(timestamp) or rollup(dimension)
        return field.for_(subquery_field).as_(field_alias)

    # Need to copy the metrics if there are references so that the `get_sql` monkey patch does not conflict
    definition = copy.deepcopy(field.definition)
    # case #2: complex blender fields
    return definition.as_(field_alias)


def _perform_join_operations(dimensions, base_query, base_field_map, join_queries, join_field_maps):
    blender_query = Query.from_(base_query, immutable=False)
    for join_sql, join_field_map in zip(join_queries, join_field_maps):
        if join_sql is None:
            continue

        criteria = _blender_join_criteria(
            base_query, join_sql, dimensions, base_field_map, join_field_map
        )

        # In most cases there are dimensions to join the two data blending queries on, but if there are none, then
        # instead of doing a join, add the data blending query to the from clause
        blender_query = (
            blender_query.from_(join_sql)  # <-- no dimensions mapped
            if criteria is None
            else blender_query.join(join_sql, JoinType.left).on(
                criteria
            )  # <-- mapped dimensions
        )

    return blender_query


def _blend_query(dimensions, metrics, orders, field_maps, queries):

    # Remove None's in the beginning and take the first non-empty query as base query
    for i, query in enumerate(queries):
        if query:
            base_query, *join_queries = queries[i:]
            base_field_map, *join_field_maps = field_maps[i:]
            break

    reference = base_query._references[0] if base_query._references else None

    if len(queries) == 1:
        # Optimization step, we don't need to do any joining as there is only one query
        blender_query = base_query
    else:
        blender_query = _perform_join_operations(dimensions, base_query, base_field_map, join_queries, join_field_maps)

        # WARNING: In order to make complex fields work, the get_sql for each field is monkey patched in. This must
        # happen here because a complex metric by definition references values selected from the dataset subqueries.

        for metric in find_dataset_fields(metrics):
            subquery_field = _get_sq_field_for_blender_field(metric, queries, field_maps, reference)
            metric.get_sql = subquery_field.get_sql

        sq_dimensions = [_get_sq_field_for_blender_field(d, queries, field_maps) for d in dimensions]
        sq_metrics = [_get_sq_field_for_blender_field(m, queries, field_maps, reference) for m in metrics]
        blender_query = blender_query.select(*sq_dimensions).select(*sq_metrics)

    for field, orientation in orders:
        if any(dimension is field for dimension in dimensions):
            # Don't add the reference type to dimensions
            orderby_field = _get_sq_field_for_blender_field(field, queries, field_maps)
        else:
            orderby_field = _get_sq_field_for_blender_field(field, queries, field_maps, reference)

        blender_query = blender_query.orderby(orderby_field, order=orientation)

    return blender_query


class DataSetBlenderQueryBuilder(DataSetQueryBuilder):
    """
    Blended dataset queries consist of widgets, dimensions, filters, orders by and references. At least one or
    more widgets is required. All others are optional.
    """

    @property
    def sql(self):
        """
        Serialize this query builder to a list of Pypika/SQL queries. This function will return one query for every
        combination of reference and rolled up dimension (including null options).

        This collects all of the metrics in each widget, dimensions, and filters and builds a corresponding pypika query
        to fetch the data.  When references are used, the base query normally produced is wrapped in an outer query and
        a query for each reference is joined based on the referenced dimension shifted.

        :return: a list of Pypika's Query subclass instances.
        """
        # First run validation for the query on all widgets
        self._validate()

        datasets, field_maps = _datasets_and_field_maps(self.dataset, self._filters)
        dimensions = self.dimensions
        metrics = find_metrics_for_widgets(self._widgets)
        metrics_aliases = {metric.alias for metric in metrics}
        dimensions_aliases = {dimension.alias for dimension in dimensions}

        # Add fields to be ordered on, to metrics if they aren't yet selected in metrics or dimensions
        for field, orientation in self.orders:
            if (
                field.alias not in metrics_aliases
                and field.alias not in dimensions_aliases
            ):
                metrics.append(field)

        dataset_metrics = find_dataset_fields(metrics)
        operations = find_operations_for_widgets(self._widgets)
        share_dimensions = find_share_dimensions(dimensions, operations)

        datasets_queries = []
        for dataset, field_map in zip(datasets, field_maps):
            datasets_queries.append(
                _build_dataset_query(
                    dataset,
                    field_map,
                    dataset_metrics,
                    dimensions,
                    self._filters,
                    self._references,
                    operations,
                    share_dimensions
                )
            )

        """
        A dataset query can yield one or more sql queries, depending on how many types of references or dimensions 
        with totals are selected. A blended dataset query must yield the same number and types of sql queries, but each
        blended together. The individual dataset queries built above will always yield the same number of sql queries, 
        so here those lists of sql queries are zipped.
        
               base   ref  totals ref+totals
        ds1 | ds1_a  ds1_b  ds1_c   ds1_d  
        ds2 | ds2_a  ds2_b  ds2_c   ds2_d  
        
        More concretely, using the diagram above as a reference, a dataset query with 1 reference and 1 totals dimension
        would yield 4 sql queries. With data blending with 1 reference and 1 totals dimension, 4 sql queries must also 
        be produced.  The following lines convert the list of rows of the table in the diagram to a list of columns.
        Each set of queries in a column are then reduced to a single data blending sql query.
        """

        # TODO: There is most likely still a bug when lots of references and total calculations get mixed.
        # Although I haven't been able to actually find a case
        per_dataset_queries_count = max(
            [len(dataset_queries) for dataset_queries in datasets_queries]
        )
        query_sets = [[] for _ in range(per_dataset_queries_count)]

        for dataset_queries in datasets_queries:
            for i, query_set in enumerate(query_sets):
                if len(dataset_queries) > i:
                    query_set.append(dataset_queries[i])

        blended_queries = []
        for queryset in query_sets:
            blended_query = _blend_query(
                dimensions, metrics, self.orders, field_maps, queryset
            )
            blended_query = self._apply_pagination(blended_query)

            if blended_query:
                blended_queries.append(blended_query)

        return blended_queries



from pypika import (
    Dialects,
    MySQLQuery,
    Table,
    enums,
    functions as fn,
    terms,
)
from pypika.terms import CustomFunction, Interval, Parameter

from . import sql_types
from .base import Database
from .type_engine import TypeEngine

_DateFormat = CustomFunction('DATE_FORMAT', ['field', 'date_format'])
_DateSub = CustomFunction('DATE_SUB', ['field', 'interval'])
_MakeDate = CustomFunction('MAKEDATE', ['year', 'day_of_year'])
_Weekday = CustomFunction('WEEKDAY', ['arg'])
_Year = CustomFunction('YEAR', ['arg'])
_Quarter = CustomFunction('QUARTER', ['arg'])
_Timestamp = CustomFunction('TIMESTAMP', ['arg'])


class DateAdd(terms.Function):
    """
    Override for the MySQL specific DateAdd function which expects an interval instead of the date part and interval
    unit e.g. DATE_ADD("date", INTERVAL 1 YEAR)
    """

    def __init__(self, field, interval_term, alias=None):
        super(DateAdd, self).__init__('DATE_ADD', field, interval_term, alias=alias)


class _CustomMySQLInterval(terms.Function):
    """
    The PyPika Interval function generates Interval SQL based on the provided time unit arguments.
    Fireant needs a more flexible Interval query string generator than this
    to mimicking the Vertica TRUNC function, so this is defined locally here.
    """
    def __init__(self, field, unit):
        super().__init__('INTERVAL', field, unit)
        self.field = field
        self.unit = unit

    def get_sql(self, **kwargs):
        field_sql = self.field.get_sql(**kwargs)
        return f'INTERVAL {field_sql} {self.unit}'


class MySQLDatabase(Database):
    """
    MySQL client that uses the PyMySQL module.
    """
    # The pypika query class to use for constructing queries
    query_cls = MySQLQuery

    def __init__(self, host='localhost', port=3306, database=None,
                 user=None, password=None, charset='utf8mb4', **kwags):
        super(MySQLDatabase, self).__init__(host, port, database, **kwags)
        self.user = user
        self.password = password
        self.charset = charset
        self.type_engine = MySQLTypeEngine()

    def _get_connection_class(self):
        # Nesting inside a function so the import does not cause issues if users have not installed the 'mysql' extra
        # when installing
        import pymysql

        class MySQLConnection(pymysql.connections.Connection):
            """
            PyMySQL has deprecated context managers in the connection class.
            To make the functionality consistent with other database drivers, we override
            the context manager to return the connection instead of the cursor object.

            This also fixes an issue where connections were not being closed in the current PyMySQL
            context manager implementation!
            https://github.com/PyMySQL/PyMySQL/issues/735
            """

            def __enter__(self):
                return self

            def __exit__(self, exc, value, traceback):
                self.close()

        return MySQLConnection

    def connect(self):
        """
        Returns a MySQL connection

        :return: pymysql Connection class
        """
        import pymysql
        connection_class = self._get_connection_class()
        return connection_class(host=self.host, port=self.port, db=self.database,
                                user=self.user, password=self.password,
                                charset=self.charset, cursorclass=pymysql.cursors.Cursor)

    def trunc_date(self, field, interval):
        if interval == 'hour':
            return _DateFormat(field, '%Y-%m-%d %H:00:00')
        elif interval == 'day':
            return _DateFormat(field, '%Y-%m-%d 00:00:00')
        elif interval == 'week':
            return _DateFormat(_DateSub(field, _CustomMySQLInterval(_Weekday(field), 'DAY')), '%Y-%m-%d 00:00:00')
        elif interval == 'month':
            return _DateFormat(field, '%Y-%m-01')
        elif interval == 'quarter':
            return _MakeDate(_Year(_Timestamp(field)), 1) \
                   + _CustomMySQLInterval(_Quarter(_Timestamp(field)), 'QUARTER') \
                   - Interval(quarters=1, dialect=Dialects.MYSQL)
        elif interval == 'year':
            return _DateFormat(field, '%Y-01-01')
        else:
            raise ValueError(f'Invalid interval provided to trunc_date method: {interval}')

    def to_char(self, definition):
        return fn.Cast(definition, enums.SqlTypes.CHAR)

    def date_add(self, field, date_part, interval):
        # adding an extra 's' as MySQL's interval doesn't work with 'year', 'week' etc, it expects a plural
        interval_term = terms.Interval(**{'{}s'.format(str(date_part)): interval, 'dialect': Dialects.MYSQL})
        return DateAdd(field, interval_term)

    def get_column_definitions(self, schema, table, connection=None):
        columns = Table('columns', schema='INFORMATION_SCHEMA')

        columns_query = MySQLQuery \
            .from_(columns) \
            .select(columns.column_name, columns.column_type) \
            .where(columns.table_schema == Parameter('%(schema)s')) \
            .where(columns.field('table_name') == Parameter('%(table)s')) \
            .distinct() \
            .orderby(columns.column_name)

        return self.fetch(str(columns_query), parameters=dict(schema=schema, table=table), connection=connection)

    def import_csv(self, table, file_path, connection=None):
        """
        Execute a query to import a file into a table using the provided connection.

        :param table: The name of a table to import data into.
        :param file_path: The path of the file to be imported.
        :param connection: (Optional) The connection to execute this query with.
        """
        import_query = MySQLQuery \
            .load(file_path) \
            .into(table)

        self.execute(str(import_query), connection=connection)

    def create_temporary_table_from_columns(self, table, columns, connection=None):
        """
        Creates a temporary table from a list of columns.

        :param table: The name of the new temporary table.
        :param columns: The columns of the new temporary table.
        :param connection: (Optional) The connection to execute this query with.
        """
        create_query = MySQLQuery \
            .create_table(table) \
            .temporary() \
            .columns(*columns)

        self.execute(str(create_query), connection=connection)

    def create_temporary_table_from_select(self, table, select_query, connection=None):
        """
        Creates a temporary table from a SELECT query.

        :param table: The name of the new temporary table.
        :param select_query: The query to be used for selecting data of an existing table for the new temporary table.
        :param connection: (Optional) The connection to execute this query with.
        """
        create_query = MySQLQuery \
            .create_table(table) \
            .temporary() \
            .as_select(select_query)

        self.execute(str(create_query), connection=connection)


class MySQLTypeEngine(TypeEngine):
    mysql_to_ansi_mapper = {
        'bit': sql_types.Char,
        'char': sql_types.Char,
        'nchar': sql_types.Char,
        'varchar': sql_types.VarChar,
        'nvarchar': sql_types.VarChar,
        'text': sql_types.Text,
        'boolean': sql_types.Boolean,
        'int': sql_types.Integer,
        'integer': sql_types.Integer,
        'year': sql_types.Integer,
        'smallint': sql_types.SmallInt,
        'tinyint': sql_types.SmallInt,
        'bigint': sql_types.BigInt,
        'decimal': sql_types.Decimal,
        'fixed': sql_types.Decimal,
        'numeric': sql_types.Numeric,
        'float': sql_types.Float,
        'real': sql_types.DoublePrecision,
        'double': sql_types.DoublePrecision,
        'date': sql_types.Date,
        'time': sql_types.Time,
        'datetime': sql_types.DateTime,
        'timestamp': sql_types.Timestamp,
    }

    ansi_to_mysql_mapper = {
        'CHAR': 'char',
        'VARCHAR': 'varchar',
        'TEXT': 'text',
        'BOOLEAN': 'boolean',
        'INTEGER': 'integer',
        'SMALLINT': 'smallint',
        'BIGINT': 'bigint',
        'DECIMAL': 'decimal',
        'NUMERIC': 'numeric',
        'FLOAT': 'float',
        'REAL': 'real',
        'DOUBLEPRECISION': 'real',
        'DATE': 'date',
        'TIME': 'time',
        'TIMESTAMP': 'timestamp',
    }

    def __init__(self):
        super(MySQLTypeEngine, self).__init__(self.mysql_to_ansi_mapper, self.ansi_to_mysql_mapper)



from fireant.dataset.fields import Field
from fireant.dataset.klass import DataSet
from fireant.queries.builder import (
    DataSetBlenderQueryBuilder,
    DimensionChoicesQueryBuilder,
)
from fireant.utils import (
    deepcopy,
    immutable,
    ordered_distinct_list_by_attr,
)


def _wrap_dataset_fields(dataset):
    if isinstance(dataset, DataSetBlender):
        return dataset.fields

    wrapped_fields = []
    for field in dataset.fields:
        wrapped_field = _wrap_field(dataset, field)

        wrapped_fields.append(wrapped_field)

    return wrapped_fields


def _wrap_field(dataset, field):
    wrapped_field = Field(
        alias=field.alias,
        definition=field,
        data_type=field.data_type,
        label=field.label,
        hint_table=field.hint_table,
        prefix=field.prefix,
        suffix=field.suffix,
        thousands=field.thousands,
        precision=field.precision,
        hyperlink_template=field.hyperlink_template,
    )

    if not field.definition.is_aggregate:
        wrapped_field.choices = DimensionChoicesBlenderQueryBuilder(dataset, field)

    return wrapped_field


class DataSetBlender:
    """
    The DataSetBlender class is the DataSet equivalent for implementing data blending, across distinct DataSet
    instances.
    """

    def __init__(self, primary_dataset, secondary_dataset, dimension_map):
        """
        Constructor for a blended dataset.  Contains all the fields to initialize the dataset.

        :param primary_dataset: (Required)
            The primary dataset, which table will be used as part of the FROM expression. This can be either a `DataSet`
            or another `DataSetBlender`, which means multiple DataSet instances can be blended by chaining together
            blenders.
        :param secondary_dataset: (Required)
            The dataset being blended. This should be a `DataSet` instance. (It might actually work with an instance of
            `DataSetBlender` as well, though.)
        :param dimension_map:
            A dict mapping up fields from the primary to the secondary dataset. This tells the Blender which fields
            can be used as dimensions in the Blender queries.
        """
        self.primary_dataset = primary_dataset
        self.secondary_dataset = secondary_dataset
        self.dimension_map = dimension_map

        # Wrap all dataset fields with another field on top so that:
        #   1. DataSetBlender doesn't share a reference to a field with a DataSet (__hash__ is used to find out which
        #      dataset the field is in - see the Field class' __hash__ method for more details)
        #   2. When complex fields are added, the `definition` attribute will always have at least one field within
        #      its object graph
        self.fields = DataSet.Fields(
            ordered_distinct_list_by_attr(
                [
                    *_wrap_dataset_fields(primary_dataset),
                    *_wrap_dataset_fields(secondary_dataset)
                ],
            )
        )

        # add query builder entry points
        self.query = DataSetBlenderQueryBuilder(self)
        self.latest = self.primary_dataset.latest
        self.annotation = None

    @property
    def return_additional_metadata(self) -> bool:
        # When using data blending, datasets are nested inside DataSetBlender objects. Additionally,
        # the primary_dataset can be a combination of datasets depending on how many datasets are being blended.
        # This helper property walks the tree to return the return_additional_metadata value from the original
        # primary dataset.
        dataset = self.primary_dataset
        while not isinstance(dataset, DataSet):
            dataset = dataset.primary_dataset

        return dataset.return_additional_metadata

    def __eq__(self, other):
        return isinstance(other, DataSetBlender) and self.fields == other.fields

    def __repr__(self):
        return "BlendedDataSet(fields=[{}])".format(
            ",".join([repr(f) for f in self.fields])
        )

    def __hash__(self):
        return hash((self.primary_dataset, self.secondary_dataset, self.fields))

    def __deepcopy__(self, memodict={}):
        for field in self.dimension_map.values():
            memodict[id(field)] = field
        return deepcopy(self, memodict)

    @property
    def table(self):
        return None

    @property
    def database(self):
        return self.primary_dataset.database

    @immutable
    def extra_fields(self, *fields):
        for field in fields:
            self.fields.add(field)

    def blend(self, other):
        """
        Returns a Data Set blender which enables to execute queries on multiple data sets and combine them.
        """
        return DataSetBlenderBuilder(self, other)


class DataSetBlenderBuilder:
    def __init__(self, primary, secondary):
        self.primary_dataset = primary
        self.secondary_dataset = secondary

    def on(self, dimension_map):
        return DataSetBlender(
            self.primary_dataset, self.secondary_dataset, dimension_map
        )

    def on_dimensions(self):
        dimension_map = {}

        for secondary_ds_field in self.secondary_dataset.fields:
            is_aggregate_field = secondary_ds_field.is_aggregate
            matches_alias_in_primary_dataset = (
                secondary_ds_field.alias in self.primary_dataset.fields
            )
            if is_aggregate_field or not matches_alias_in_primary_dataset:
                continue

            primary_ds_field = self.primary_dataset.fields[secondary_ds_field.alias]
            dimension_map[primary_ds_field] = secondary_ds_field

        return self.on(dimension_map)


class DimensionChoicesBlenderQueryBuilder(DimensionChoicesQueryBuilder):
    def filter(self, *filters, **kwargs):
        filters = [
            fltr.for_(fltr.field.definition)
            for fltr in filters
            if fltr.field.definition in self.dataset.fields
        ]
        return super().filter(*filters, **kwargs)



from fireant.dataset.filters import ComparisonOperator
from fireant.utils import alias_selector


def reference_alias(metric, reference):
    """
    Format a metric key for a reference.

    :return:
        A string that is used as the key for a reference metric.
    """
    key = metric.alias

    if reference is None:
        return key

    return '{}_{}'.format(key, reference.alias)


def reference_type_alias(metric, reference):
    """
    Format a metric key for a subquery selection in case of a reference.

    :return:
        A string that is used as the selector for a field in a blended subquery.
    """
    key = metric.alias

    if reference is None:
        return key

    return '{}_{}'.format(key, reference.reference_type.alias)


def reference_label(metric, reference):
    """
    Format a metric label for a reference.

    :return:
        A string that is used as the display value for a reference metric.
    """
    label = str.strip(metric.label or metric.alias)

    if reference is None:
        return label

    return '{} {}'.format(label, reference.label)


def reference_prefix(metric, reference):
    """
    Return the prefix for a metric displayed for a reference (or no Reference)

    :return:
        A string that is used as the prefix for a reference metric.
    """
    if reference is not None and reference.delta_percent:
        return None
    return metric.prefix


def reference_suffix(metric, reference):
    """
    Return the suffix for a metric displayed for a reference (or no Reference)

    :return:
        A string that is used as the suffix for a reference metric.
    """
    if reference is not None and reference.delta_percent:
        return '%'
    return metric.suffix


def apply_reference_filters(df, reference):
    for reference_filter in reference.filters:
        df_column_key = alias_selector(reference_alias(reference_filter.metric, reference))
        if df_column_key in df:
            column = df[df_column_key]
            dataframe_filter = ComparisonOperator.eval(column, reference_filter.operator, reference_filter.value)
            df = df.loc[dataframe_filter]

    return df



import logging
from typing import Iterable, List, Tuple, Type

import pandas as pd
from pypika.queries import QueryBuilder

from fireant.database import Database
from fireant.dataset.fields import DataType, Field
from fireant.dataset.references import calculate_delta_percent
from fireant.dataset.totals import get_totals_marker_for_dtype
from fireant.utils import alias_selector, chunks
from .finders import find_field_in_modified_field, find_totals_dimensions
from .pandas_workaround import df_subtract
from ..dataset.modifiers import RollupValue

logger = logging.getLogger(__name__)

# Passing in an empty dictionary as format option to pandas' parse_dates causes errors to be ignored instead of
# being coerced into NaT.
PANDAS_TO_DATETIME_FORMAT = {}


def fetch_data(
    database: Database,
    queries: List[Type[QueryBuilder]],
    dimensions: Iterable[Field],
    share_dimensions: Iterable[Field] = (),
    reference_groups=(),
) -> Tuple[int, pd.DataFrame]:
    queries = [str(query) for query in queries]

    # Indicate which dimensions need to be parsed as date types
    # For this we create a dictionary with the dimension alias as key and PANDAS_TO_DATETIME_FORMAT as value
    pandas_parse_dates = {}
    for dimension in dimensions:
        unmodified_dimension = find_field_in_modified_field(dimension)
        if unmodified_dimension.data_type == DataType.date:
            pandas_parse_dates[alias_selector(unmodified_dimension.alias)] = PANDAS_TO_DATETIME_FORMAT

    results = database.fetch_dataframes(*queries, parse_dates=pandas_parse_dates)
    max_rows_returned = max([len(x) for x in results], default=0)
    logger.info('max_rows_returned', extra={'row_count': max_rows_returned, 'database': str(database)})

    return max_rows_returned, reduce_result_set(results, reference_groups, dimensions, share_dimensions)


def reduce_result_set(
    results: Iterable[pd.DataFrame],
    reference_groups,
    dimensions: Iterable[Field],
    share_dimensions: Iterable[Field],
):
    """
    Reduces the result sets from individual queries into a single data frame. This effectively joins sets of references
    and concatenates the sets of totals.

    :param results: A list of data frame
    :param reference_groups: A list of groups of references (grouped by interval such as WoW, etc)
    :param dimensions: A list of dimensions, used for setting the index on the result data frame.
    :param share_dimensions: A list of dimensions from which the totals are used for calculating share operations.
    :return:
    """
    # One result group for each rolled up dimension. Groups contain one member plus one for each reference type used.
    result_groups = chunks(results, 1 + len(reference_groups))

    dimension_keys = [alias_selector(d.alias) for d in dimensions]
    totals_dimension_keys = [
        alias_selector(d.alias)
        for d in find_totals_dimensions(dimensions, share_dimensions)
    ]
    dimension_dtypes = result_groups[0][0][dimension_keys].dtypes

    # Reduce each group to one data frame per rolled up dimension
    group_data_frames = []
    for i, result_group in enumerate(result_groups):
        if dimension_keys:
            result_group = [result.set_index(dimension_keys) for result in result_group]

        base_df = result_group[0]
        reference_dfs = [
            _make_reference_data_frame(base_df, result, reference)
            for result, reference_group in zip(result_group[1:], reference_groups)
            for reference in reference_group
        ]

        # Merge the different reference dataframes into 1
        merged_df = base_df
        for reference_df in reference_dfs:
            merged_df = pd.merge(
                merged_df, reference_df, how="outer", left_index=True, right_index=True, suffixes=('', '_delete')
            )
            # Drop duplicate columns from the merge (indicate by suffix _delete)
            merged_df.drop(merged_df.filter(regex='_delete$').columns.tolist(), axis=1, inplace=True)

        # If there are rolled up dimensions in this result set then replace the NaNs for that dimension value with a
        # marker to indicate totals.
        # The data frames will be ordered so that the first group will contain the data without any rolled up
        # dimensions, then followed by the groups with them, ordered by the last rollup dimension first.
        if totals_dimension_keys[:i]:
            merged_df = _replace_rollup_constants_for_totals_markers(merged_df, dimension_dtypes)

        group_data_frames.append(merged_df)

    return pd.concat(group_data_frames, sort=False).sort_index(na_position="first")


def _replace_rollup_constants_for_totals_markers(data_frame, dtypes):
    # some things are just easier to do without an index. Reset it temporarily to replace Rollup constants with the
    # rollup marker values
    index_names = data_frame.index.names
    data_frame.reset_index(inplace=True)

    for dimension_key, dtype in dtypes.items():
        data_frame[dimension_key] = data_frame[dimension_key].replace(
            RollupValue.CONSTANT, get_totals_marker_for_dtype(dtype)
        )

    return data_frame.set_index(index_names)


def _make_reference_data_frame(base_df, ref_df, reference):
    """
    This applies the reference metrics to the data frame given the base data frame and the reference data frame.

    When a reference is selected as a delta or a delta percentage, the calculation is performed here. Otherwise, the
    reference data frame is returned.

    :param base_df:
    :param ref_df:
    :param reference:
    :return:
    """
    metric_column_indices = [
        i for i, column in enumerate(ref_df.columns) if column not in base_df.columns
    ]
    ref_columns = [ref_df.columns[i] for i in metric_column_indices]

    if not (reference.delta or reference.delta_percent):
        return ref_df[ref_columns]

    original_ref_df = ref_df

    base_columns = [base_df.columns[i] for i in metric_column_indices]

    # Select just the metric columns from the DF and rename them with the reference key as a suffix
    base_df, ref_df = base_df[base_columns].copy(), ref_df[ref_columns].copy()
    # Both data frame columns are renamed in order to perform the calculation below.
    base_df.columns = ref_df.columns = [
        column.replace(reference.reference_type.alias, reference.alias)
        for column in ref_columns
    ]

    ref_delta_df = df_subtract(base_df, ref_df, fill_value=0)

    if reference.delta_percent:
        ref_delta_df = calculate_delta_percent(ref_df, ref_delta_df)

    # Add original reference values back to df
    ref_delta_df[ref_columns] = original_ref_df[ref_columns]
    # Alternative for the above line but with copying:
    # ref_delta_df = ref_delta_df.join(original_ref_df[ref_columns])

    return ref_delta_df



from pypika import JoinType


class Join(object):
    """
    The `Join` class represents a traditional table join in the `DataSet` API. For joining sub-queries see the
    `DataSetJoin` class.
    """

    def __init__(self, table, criterion, join_type=JoinType.inner):
        """
        :param table:
            A PyPika Table instance.
        :param criterion:
            A PyPika boolean expression.
        :param join_type:
            The join type (e.g. inner, left, right). See JoinType enum for more info.
        """
        self.table = table
        self.criterion = criterion
        self.join_type = join_type

    def __repr__(self):
        return "{type} JOIN {table} ON {criterion}".format(
            type=self.join_type, table=self.table, criterion=self.criterion
        )

    def __gt__(self, other):
        return self.table._table_name < other.table._table_name



class ANSIType:
    """
    Represents an ANSI data type.
    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, ANSIType):
            return str(self) == str(other)
        return self.name == str(other)

    def __str__(self):
        return self.name

    @property
    def params(self):
        return ''


class Char(ANSIType):
    """
    Represents the ANSI type CHAR.
    """
    def __init__(self, length=None):
        super(Char, self).__init__(name='CHAR')
        self.length = length

    def __str__(self):
        return '{name}{params}'.format(
              name=self.name,
              params=self.params
        )

    @property
    def params(self):
        return '({length})'.format(length=self.length) if self.length else ''


class VarChar(ANSIType):
    """
    Represents the ANSI type VARCHAR.
    """
    def __init__(self, length=None):
        super(VarChar, self).__init__(name='VARCHAR')
        self.length = length

    def __str__(self):
        return '{name}{params}'.format(
              name=self.name,
              params=self.params
        )

    @property
    def params(self):
        return '({length})'.format(length=self.length) if self.length else ''


class Text(ANSIType):
    """
    Represents the ANSI type TEXT.
    """
    def __init__(self):
        super(Text, self).__init__(name='TEXT')


class Boolean(ANSIType):
    """
    Represents the ANSI type BOOLEAN.
    """
    def __init__(self):
        super(Boolean, self).__init__(name='BOOLEAN')


class Integer(ANSIType):
    """
    Represents the ANSI type INTEGER.
    """
    def __init__(self):
        super(Integer, self).__init__(name='INTEGER')


class SmallInt(ANSIType):
    """
    Represents the ANSI type SMALLINT.
    """
    def __init__(self):
        super(SmallInt, self).__init__(name='SMALLINT')


class BigInt(ANSIType):
    """
    Represents the ANSI type BIGINT.
    """
    def __init__(self):
        super(BigInt, self).__init__(name='BIGINT')


class Decimal(ANSIType):
    """
    Represents the ANSI type DECIMAL.
    """
    def __init__(self, precision=None, scale=None):
        super(Decimal, self).__init__(name='DECIMAL')
        self.precision = precision
        self.scale = scale

    def __str__(self):
        return '{name}{params}'.format(
              name=self.name,
              params=self.params
        )

    @property
    def params(self):
        if not self.precision or not self.scale:
            return ''

        return '({precision},{scale})'.format(
              precision=self.precision,
              scale=self.scale,
        )


class Numeric(ANSIType):
    """
    Represents the ANSI type NUMERIC.
    """
    def __init__(self, precision, scale):
        super(Numeric, self).__init__(name='NUMERIC')
        self.precision = precision
        self.scale = scale

    def __str__(self):
        return '{name}{params}'.format(
              name=self.name,
              params=self.params
        )

    @property
    def params(self):
        if not self.precision or not self.scale:
            return ''

        return '({precision},{scale})'.format(
              precision=self.precision,
              scale=self.scale,
        )


class Float(ANSIType):
    """
    Represents the ANSI type FLOAT.
    """
    def __init__(self, precision):
        super(Float, self).__init__(name='FLOAT')
        self.precision = precision

    def __str__(self):
        return '{name}{params}'.format(
              name=self.name,
              params=self.params
        )

    @property
    def params(self):
        return '({precision})'.format(precision=self.precision) if self.precision else ''


class Real(ANSIType):
    """
    Represents the ANSI type REAL.
    """
    def __init__(self):
        super(Real, self).__init__(name='REAL')


class DoublePrecision(ANSIType):
    """
    Represents the ANSI type DOUBLEPRECISION.
    """
    def __init__(self):
        super(DoublePrecision, self).__init__(name='DOUBLEPRECISION')


class Date(ANSIType):
    """
    Represents the ANSI type DATE.
    """
    def __init__(self):
        super(Date, self).__init__(name='DATE')


class Time(ANSIType):
    """
    Represents the ANSI type TIME.
    """
    def __init__(self):
        super(Time, self).__init__(name='TIME')


class DateTime(ANSIType):
    """
    Represents the ANSI type DATETIME.
    """
    def __init__(self):
        super(DateTime, self).__init__(name='DATETIME')


class Timestamp(ANSIType):
    """
    Represents the ANSI type TIMESTAMP.
    """
    def __init__(self):
        super(Timestamp, self).__init__(name='TIMESTAMP')





from .dataset_blender_query_builder import DataSetBlenderQueryBuilder
from .dataset_query_builder import DataSetQueryBuilder
from .dimension_choices_query_builder import DimensionChoicesQueryBuilder
from .dimension_latest_query_builder import DimensionLatestQueryBuilder
from .query_builder import (
    QueryBuilder,
    QueryException,
    add_hints,
    get_column_names,
)



from pypika.queries import Column as PypikaColumn


class Column:
    """
    Represents an abstract database column.
    """
    def __init__(self, column_name, column_type):
        self.name = column_name
        self.type = column_type

    def __eq__(self, other):
        return str(self) == str(other)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '{name} {type}'.format(
              name=self.name,
              type=str(self.type),
        )

    def to_database_column(self, type_engine):
        """
        Returns a database specific column representation. The provided type engine is used to match the columns type
        to its database specific representation.

        :param type_engine: The type engine for converting the column's type.
        :return: A database specific representation of the column.
        """
        return '{name} {type}'.format(
              name=self.name,
              type=type_engine.from_ansi(self.type),
        )

    def to_pypika_column(self, type_engine):
        """
        Returns the column as a Pypika Column instance.

        :param type_engine: The type engine for converting the column's type.
        :return: A Pypika Column instance.
        """
        return PypikaColumn(
              column_name=self.name,
              column_type=type_engine.from_ansi(self.type),
        )


def make_columns(db_columns, db_type_engine):
    """
    Creates a list of Column instances with ANSI types from db columns. The function expects
    that the provided type engine is able match the provided db column types.

    :param db_columns: A list of column definitions.
    :param db_type_engine: A type engine to convert column types to ansi types.
    :return: A list of Column instances with ANSI types.
    """
    columns = []
    for column_definition in db_columns:
        column_name = column_definition[0]
        column_type = db_type_engine.to_ansi(column_definition[1])
        columns.append(Column(column_name, column_type))

    return columns


class ColumnsTransformer:
    @staticmethod
    def database_to_ansi(db_columns, db):
        """
        Transforms database specific columns to a list of Column instances.

        :param db_columns: The list of database columns.
        :param db: The database instance.
        :return: The columns of the secondary tables as a list of Column instances.
        """
        return make_columns(db_columns, db.type_engine)

    @staticmethod
    def ansi_to_database(columns, db):
        """
        Transforms a list of Column instances to database specific Pypika-Columns.

        :param columns: The list of Column instances.
        :param db: The database instance.
        :return: A list of Pypika-Columns with the transformed database types.
        """
        return [column.to_pypika_column(db.type_engine) for column in columns]

    @staticmethod
    def database_to_database(columns, from_db, to_db):
        """
        Transforms columns from one database to Pypika-Columns of another database.

        :param columns: The list of columns from one database.
        :param from_db: The origin database instance.
        :param to_db: The target database instance.
        :return: A list of Pypika-Columns with the target database types.
        """
        ansi_columns = ColumnsTransformer.database_to_ansi(columns, from_db)
        return ColumnsTransformer.ansi_to_database(ansi_columns, to_db)




import pandas as pd


def _reindex_with_nans(df, idx, fill_value=None):
    missing = pd.DataFrame(index=idx.symmetric_difference(df.index), columns=df.columns)
    if fill_value is not None:
        missing = missing.fillna(fill_value)
    return df.append(missing).reindex(idx)


def _reindex_deduplicate(left, right, fill_value=None):
    combined_index = left.index.append(right.index)
    dededuplicated_index = combined_index[~combined_index.duplicated()]
    left_reindex, right_reindex = [
        _reindex_with_nans(df, dededuplicated_index, fill_value=fill_value)
        for df in (left, right)
    ]
    return left_reindex, right_reindex


def df_subtract(left, right, fill_value=None):
    left_reindex, right_reindex = _reindex_deduplicate(
        left, right, fill_value=fill_value
    )
    return left_reindex.subtract(right_reindex)
